"""Tests for coBib's ShowCommand."""
# pylint: disable=no-self-use,unused-argument

from __future__ import annotations

from argparse import Namespace
from io import StringIO
from itertools import zip_longest
from typing import TYPE_CHECKING, Any, List, Optional, Type

import pytest

from cobib.commands import ShowCommand
from cobib.config import Event

from .. import get_resource
from ..tui.tui_test import TUITest
from .command_test import CommandTest

if TYPE_CHECKING:
    import cobib.commands


class TestShowCommand(CommandTest, TUITest):
    """Tests for coBib's ShowCommand."""

    def get_command(self) -> Type[cobib.commands.base_command.Command]:
        # noqa: D102
        return ShowCommand

    def _assert(self, output: List[str]) -> None:
        """Common assertion utility method.

        Args:
            output: the actual output of the command.
        """
        with open(get_resource("example_literature.bib"), "r", encoding="utf-8") as expected:
            # we use zip_longest to ensure that we don't have more than we expect
            for line, truth in zip_longest(output, expected):
                if not line:
                    continue
                assert line == truth.strip("\n")

    def test_command(self, setup: Any) -> None:  # type: ignore
        """Test the command itself.

        Args:
            setup: the `tests.commands.command_test.CommandTest.setup` fixture.
        """
        # redirect output of show to string
        file = StringIO()
        ShowCommand().execute(["einstein"], out=file)
        self._assert(file.getvalue().split("\n"))

    def test_warning_missing_label(self, setup: Any, caplog: pytest.LogCaptureFixture) -> None:
        """Test warning for missing label.

        Args:
            setup: the `tests.commands.command_test.CommandTest.setup` fixture.
            caplog: the built-in pytest fixture.
        """
        ShowCommand().execute(["dummy"])
        assert (
            "cobib.commands.show",
            40,
            "No entry with the label 'dummy' could be found.",
        ) in caplog.record_tuples

    def test_cmdline(
        self, setup: Any, monkeypatch: pytest.MonkeyPatch, capsys: pytest.CaptureFixture[str]
    ) -> None:
        """Test the command-line access of the command.

        Args:
            setup: the `tests.commands.command_test.CommandTest.setup` fixture.
            monkeypatch: the built-in pytest fixture.
            capsys: the built-in pytest fixture.
        """
        self.run_module(monkeypatch, "main", ["cobib", "show", "einstein"])
        self._assert(capsys.readouterr().out.strip().split("\n"))

    @pytest.mark.parametrize(
        ["select", "keys"],
        [
            [False, "\n"],
            [True, "v\n"],
        ],
    )
    def test_tui(self, setup: Any, select: bool, keys: str) -> None:
        """Test the TUI access of the command.

        Args:
            setup: the `tests.commands.command_test.CommandTest.setup` fixture.
            select: whether to use the TUI selection.
            keys: the string of keys to pass to the TUI.
        """

        def assertion(screen, logs, **kwargs):  # type: ignore
            expected_screen = [
                "@misc{knuthwebsite,",
                "author = {Donald Knuth},",
                "title = {Knuth: Computers and Typesetting},",
                r"url = {http://www-cs-faculty.stanford.edu/\~{}uno/abcde.html}",
                "}",
            ]
            for line, truth in zip(expected_screen, screen.display[1:]):
                assert line == truth.strip()

            assert "knuthwebsite" in screen.display[0]

            expected_log = [
                ("cobib.commands.show", 10, "Show command triggered from TUI."),
                ("cobib.commands.show", 10, "Clearing current buffer contents."),
                ("cobib.commands.show", 10, "Starting Show command."),
                ("cobib.commands.show", 10, "Populating buffer with ShowCommand result."),
            ]
            if kwargs.get("select", False):
                expected_log.insert(
                    3,
                    (
                        "cobib.commands.show",
                        10,
                        "Current entry is selected. Applying highlighting.",
                    ),
                )
            assert [log for log in logs if log[0] == "cobib.commands.show"] == expected_log

            label_color = "magenta" if kwargs.get("select", False) else "cyan"
            assert [c.bg for c in screen.buffer[1].values()][0:6] == ["cyan"] * 6
            assert [c.bg for c in screen.buffer[1].values()][6:18] == [label_color] * 12
            assert [c.bg for c in screen.buffer[1].values()][18:] == ["cyan"] * 62

        self.run_tui(keys, assertion, {"select": select})

    def test_event_pre_show_command(self, setup: Any) -> None:
        """Tests the PreShowCommand event."""

        @Event.PreShowCommand.subscribe
        def hook(largs: Namespace) -> None:
            largs.label = "einstein"

        assert Event.PreShowCommand.validate()

        file = StringIO()
        ShowCommand().execute(["knuthwebsite"], out=file)
        self._assert(file.getvalue().split("\n"))

    def test_event_post_show_command(self, setup: Any) -> None:
        """Tests the PostShowCommand event."""

        @Event.PostShowCommand.subscribe
        def hook(entry_str: str) -> Optional[str]:
            return "Hello world!"

        assert Event.PostShowCommand.validate()

        file = StringIO()
        ShowCommand().execute(["knuthwebsite"], out=file)
        assert file.getvalue() == "Hello world!\n"



"""coBib's auxiliary in-/output buffers.

coBib's TUI manages in- and output through auxiliary buffers.
These take care of everything from simple text buffering to highlighting conversion from ANSI
color codes to `curses` color pairs.
"""

from __future__ import annotations

import curses
import logging
import re
import textwrap
from typing import TYPE_CHECKING, Dict, List, Optional, Tuple, Union

LOGGER = logging.getLogger(__name__)

if TYPE_CHECKING:
    import cobib.tui


class TextBuffer:
    """An auxiliary output-buffer.

    This class implements all necessary methods to make it a suitable argument in a scenario like:
    ```python
    print("some text", file=TextBuffer())
    ```

    This means, it can buffer text which will be subsequently displayed in a `curses.pad`.
    To this extent this class also implements an ANSI color code to `curses` color pair conversion
    as part of its `view` method.

    Finally, the `popup` method can be leveraged by `cobib.tui.tui.TUI` for an easy way to display
    an overlaid frame within the TUI.
    """

    INDENT: str = "↪"
    """The string to be used for line continuation."""

    def __init__(self) -> None:
        """Initializes the TextBuffer object."""
        self.lines: List[str] = []
        """The actual text buffer."""
        self.height: int = 0
        """The height of the buffer contents (i.e. the number of lines)."""
        self.width: int = 0
        """The width of the buffer contents (i.e. the maximum line width)."""
        self.wrapped: bool = False
        """Whether the buffer contents have been wrappped to a certain width."""
        self.ansi_map: Optional[Dict[str, int]] = None
        """An optional dictionary mapping ANSI color codes to `curses` color pair indices."""

    def write(self, string: str) -> None:
        """Writes a non-empty string into the buffer.

        Args:
            string: the string to append to the buffer.
        """
        if string.strip():
            # only handle non-empty strings
            LOGGER.debug("Appending string to text buffer: %s", string)
            self.lines.append(string)
            self.height = len(self.lines)
            self.width = max(self.width, len(string))

    def replace(self, lines: Union[int, range, List[int]], old_str: str, new_str: str) -> None:
        """Replaces the old string with the new in the given lines.

        Args:
            lines: index or indices on which lines to do the replacement.
            old_str: old string to be replaced.
            new_str: new string to be inserted.
        """
        LOGGER.debug('Replacing "%s" with "%s" in lines %s', old_str, new_str, lines)
        if isinstance(lines, int):
            lines = [lines]
        for idx in lines:
            self.lines[idx] = re.sub(old_str, new_str, self.lines[idx])

    def flush(self) -> None:
        """Compatibility method. It has no effect."""

    def close(self) -> None:
        """Compatibility method. It has no effect."""

    def clear(self) -> None:
        """Clears the buffer."""
        LOGGER.debug("Clearing text buffer.")
        self.lines = []
        self.height = 0
        self.width = 0
        self.wrapped = False

    def split(self) -> None:
        """Splits the buffer lines at literal line breaks."""
        copy = self.lines.copy()
        self.lines = []
        self.width = 0
        for line in copy:
            for string in line.split("\n"):
                self.lines.append(string)
                self.width = max(self.width, len(string))
        self.height = len(self.lines)

    def wrap(self, width: int, label_column: bool = True) -> None:
        """Wraps the text in the buffer at a given width.

        If `label_column` is set to `True` (the default), this method will determine the width of
        the column containing the entry labels.
        This is done by attempting to split each line around two consecutive spaces ("␣␣") and
        determining the width of the first field.
        This assumes that the `cobib.commands.list.ListCommand` *always* outputs the entry labels as
        the first column.
        This label column length is then used on indented lines to achieve properly wrapped table
        columns in the output.

        On the other hand, if this value is set to `False`, a default initial column width of `2` is
        used.

        Args:
            width: maximum text width for each line in the buffer.
            label_column: whether to determine a label column width.
        """
        copy = self.lines.copy()
        self.lines = []
        self.width = 0
        if self.wrapped:
            LOGGER.debug("Unwrapping text buffer.")
            for line in copy:
                # unwrap instead
                if line.startswith(TextBuffer.INDENT):
                    # Note: insert single space when joining lines and strip INDENT symbol
                    self.lines[-1] += " " + line[1:].strip()
                else:
                    self.lines.append(line)
                self.width = max(self.width, len(self.lines[-1]))
        else:
            LOGGER.debug("Wrapping text buffer.")
            if label_column:
                # first, determine width of label column
                label_len = 0
                for line in copy:
                    label = line.split("  ")[0]
                    label_len = max(len(label) + 1, label_len)
                    if label_len > width:
                        LOGGER.debug(
                            "Label column width would exceed actual width. Continuing"
                            "without a label column."
                        )
                        label_len = 1
                        break
                LOGGER.debug("Label column width determined to be %d", label_len)
            else:
                label_len = 2
                LOGGER.debug("Usig default label column width %d", label_len)
            for line in copy:
                # then wrap lines with subsequent indents matched to first column width
                for string in textwrap.wrap(
                    line, width=width - 1, subsequent_indent=TextBuffer.INDENT + " " * label_len
                ):
                    self.lines.append(string)
                self.width = width
        self.height = len(self.lines)
        self.wrapped = not self.wrapped

    # pylint: disable=too-many-arguments
    def view(
        self,
        pad: "curses.window",  # type: ignore
        smaxrow: int,
        smaxcol: int,
        pminrow: int = 0,
        pmincol: int = 0,
        sminrow: int = 1,
        smincol: int = 0,
        ansi_map: Optional[Dict[str, int]] = None,
        background: Optional[int] = None,
        box: bool = False,
    ) -> None:
        """Views the buffer contents in the provided `curses.window`.

        Args:
            pad: a re-sizable curses window (aka a pad).
            smaxrow: the available height for the pad.
            smaxcol: the available width for the pad.
            pminrow: the upper display position in the pad.
            pmincol: the left display position in the pad.
            sminrow: the upper display position of the pad.
            smincol: the left display position of the pad.
            ansi_map: optional, dictionary mapping ANSI codes to curses color pairs.
            background: the curses color pair number with which to highlight the window.
            box: whether to print a frame around the window.
        """
        # a regex to detect ANSI color codes
        ansi_regex = re.compile(r"(\x1b\[(\d+)[;]*(\d+)*m)")
        if ansi_map:
            LOGGER.debug("Interpreting ANSI color codes on the fly.")
            self.ansi_map = ansi_map
        # first clear pad
        LOGGER.debug("Clearing curses pad.")
        pad.erase()
        LOGGER.debug("Pad view parameters:")
        LOGGER.debug((pminrow, pmincol, sminrow, smincol, smaxrow, smaxcol))
        pad.refresh(pminrow, pmincol, sminrow, smincol, smaxrow, smaxcol)
        # then resize
        LOGGER.debug("Adjusting pad size.")
        # NOTE The +1 added onto the height accounts for some weird offset in the curses pad.
        pad.resize(self.height + (2 if box else 1), max(self.width, smaxcol + (0 if box else 1)))
        # and populate
        for row, line in enumerate(self.lines):
            start, end, color = -1, -1, -1
            # This list will store the spanned regions for each ANSI color pair.
            # Its entries will be (curses color pair number, start, end).
            color_spans: List[Tuple[int, int, int]] = []
            if self.ansi_map:
                LOGGER.debug("Applying ANSI color map.")
                # The index below is used to keep track of the oldest (i.e. lowest in index) color
                # span which has not been closed yet. This means that we work with the assumption
                # that ANSI color codes always occur in pairs (even though a single \x1b[0m sequence
                # could terminate multiple open spans).
                lowest_incomplete_color_span = 0

                # In order to correctly trim the old line while piecing together the new one we need
                # to keep track of the previously encountered end position.
                prev_end = 0
                new_line = ""

                # iterate over all ANSI color code matches on the current line
                for match in ansi_regex.finditer(line):
                    # add everything preceding the current match to the new line
                    new_line += line[: match.start() - prev_end]
                    # trim the current line from everything in front of the match and itself
                    line = line[match.end() - prev_end :]
                    # store the current end position to use as an offset on the next match
                    prev_end = match.end()

                    # handle ANSI color code
                    if match[0] == "\x1b[0m":
                        # a closing sequence completes the lowest incomplete color span
                        # Note: as mentioned above, we work under the assumption that all ANSI color
                        # codes occur in pairs!
                        color_spans[lowest_incomplete_color_span] = (
                            color_spans[lowest_incomplete_color_span][0],
                            color_spans[lowest_incomplete_color_span][1],
                            len(new_line),
                        )
                        lowest_incomplete_color_span += 1
                    else:
                        # else we create a new color span
                        color_spans.append((self.ansi_map[match[0]], len(new_line), -1))

                # anything left from the original line needs to be added to the new one
                new_line += line
                # and finally the new line replaces the original one
                line = new_line

            pad.addstr(row + (1 if box else 0), (1 if box else 0), line)
            for color, start, end in sorted(color_spans):
                pad.chgat(row, start, end - start, curses.color_pair(color))
        # apply decorations
        if background is not None:
            # setting background color
            pad.bkgd(" ", curses.color_pair(background + 1))
        if box:
            pad.box()
        LOGGER.debug("Viewing curses pad.")
        pad.refresh(pminrow, pmincol, sminrow, smincol, smaxrow, smaxcol)

    def popup(self, tui: cobib.tui.TUI, background: Optional[int] = None) -> None:
        """Displays the buffer in a popup window.

        Args:
            tui: the TUI instance in which the popup is displayed.
            background: the curses color pair number with which to highlight the window.
        """
        LOGGER.debug("Create popup window.")
        popup_win = curses.newpad(self.height + 2, tui.width)
        # computing height offset
        height_offset = tui.height - self.height - 4
        # wrap buffer prior to viewing
        self.wrap(tui.width, label_column=False)
        # view popup window
        self.view(
            popup_win,
            min(height_offset + self.height + 2, tui.height - 1),
            tui.width,
            sminrow=height_offset,
            background=background,
            box=True,
        )

        key = 0
        # loop until quit by user
        while key not in (27, ord("q")):  # exit on ESC
            if key != 0:
                # do not print warning on initial run (key == 0)
                tui.prompt_print('To quit the popup window, press "q".')
            key = tui.prompt.getch()

        # close popup window
        popup_win.clear()
        tui.resize_handler(None, None)


class InputBuffer:
    """An auxiliary output-buffer.

    This class implements all necessary methods to make it a suitable replacement for `sys.stdin`.
    This is used when a command-line implementation of a command queries a user for input on `stdin`
    and allows the TUI to react accordingly.
    """

    def __init__(self, buffer: TextBuffer, tui: cobib.tui.TUI) -> None:
        """Initializes the InputBuffer object.

        Args:
            buffer: the current `sys.stdout` replacement.
            tui: the current TUI instance.
        """
        self.buffer = buffer
        """The `TextBuffer` replaces `sys.stdout`."""
        self.tui = tui
        """The `cobib.tui.tui.TUI` instance whose `cobib.tui.tui.TUI.prompt_handler` to use."""

    def readline(self) -> str:
        """Reads an input from the user mimicking `sys.stdin`."""
        self.buffer.split()
        LOGGER.debug("Create popup window.")
        popup_win = curses.newpad(self.buffer.height + 2, self.tui.width)
        # view popup window
        height_offset = self.tui.height - self.buffer.height - 4
        # wrap buffer prior to viewing
        self.buffer.wrap(self.tui.width, label_column=False)
        # view buffer
        self.buffer.view(
            popup_win,
            height_offset + self.buffer.height + 2,
            self.tui.width,
            sminrow=height_offset,
            box=True,
        )

        user_input: str = self.tui.prompt_handler(None, symbol="> ")

        # close popup window
        popup_win.clear()
        self.buffer.clear()
        self.tui.resize_handler(None, None)

        return user_input



"""coBib's TUI module.

This module implements coBib's curses-based TUI.
The code is split into several submodules:
* `cobib.tui.buffer`: implements a general `TextBuffer` and `InputBuffer` to handle in- and output.
* `cobib.tui.frame`: implements the `Frame` class which tightly couples a `TextBuffer` to a
  `curses.pad` and is used by the `TUI` as the main interaction window.
* `cobib.tui.state`: gathers the stateful-information of the `TUI` in a simple object.
* `cobib.tui.tui.TUI`: implements the actual TUI.
"""

import curses

from .buffer import InputBuffer as InputBuffer
from .buffer import TextBuffer as TextBuffer
from .frame import Frame as Frame
from .state import Mode as Mode
from .state import State as State
from .tui import TUI as TUI


def tui() -> None:
    """The main executable for the curses-TUI.

    This method is the entry-point for coBib's TUI.
    It simply passes the `cobib.tui.tui.TUI` initializer to the `curses.wrapper` method which takes
    care of all the rest.
    """
    curses.wrapper(TUI)  # pragma: no cover



"""coBib's TUI State object.

This object gathers all the stateful information of the TUI.
It provides utility methods to `reset` the State to it's starting point and an `update` method to
set the State to a previously stored version.
"""

from __future__ import annotations

import logging
from enum import Enum
from typing import List

from cobib.config import config

LOGGER = logging.getLogger(__name__)


class Mode(Enum):
    """An `Enum` listing the possible modes of the main window."""

    LIST = "list"
    """This mode is the default one and contains output from a `cobib.commands.list.ListCommand`."""
    SHOW = "show"
    """This mode contains output from a `cobib.commands.show.ShowCommand`."""
    SEARCH = "search"
    """This mode contains output from a `cobib.commands.search.SearchCommand`."""


class State:
    """An object to track the stateful parameters of coBib's TUI.

    This object is used to store all stateful parameters of coBib's TUI and simplify the handling of
    these parameters across the TUI and Frame objects.
    """

    def __init__(self) -> None:
        """Initializes the State object."""
        LOGGER.debug("Initializing the State")
        self.top_line: int = 0
        """The line number at the top of the visible window."""
        self.left_edge: int = 0
        """The column number at the left edge of the visible window."""
        self.current_line: int = 0
        """The currently selected line number."""
        self.previous_line: int = -1
        """The previous line number. This is used when switching back to a previous `Mode`."""

        self.mode: str = Mode.LIST.value
        """The window's current `Mode`."""
        self.inactive_commands: List[str] = []
        """A list of disabled commands. For example the *Search* Mode disables the `Add`, `Filter`
        and `Sort` commands."""
        self.topstatus: str = ""
        """The contents of the top statusbar."""

        # these cannot be set yet, because the `config` has not been fully populated at the time of
        # creation of the STATE singleton
        self.list_args: List[str] = []
        """The current arguments for the `cobib.commands.list.ListCommand` populating the default
        *List* Mode. This variable is crucial for the handling of the `Sort` and `Filter`
        commands."""

    def reset(self) -> None:
        """Resets the state to its original starting point."""
        LOGGER.debug("Resetting the State")
        self.top_line = 0
        self.left_edge = 0
        self.current_line = 0
        self.previous_line = -1

        self.mode = Mode.LIST.value
        self.inactive_commands = []
        self.topstatus = ""

        self.initialize()

    def initialize(self) -> None:
        """Initializes the configuration-dependent settings."""
        self.list_args = config.tui.default_list_args
        if config.tui.reverse_order:
            self.list_args += ["-r"]

    def update(self, state: State) -> None:
        """Update from another State.

        This overwrites this State with the contents of another one.

        Args:
            state: another (deep-copied) state.
        """
        self.top_line = state.top_line
        self.left_edge = state.left_edge
        self.current_line = state.current_line
        self.previous_line = state.previous_line

        self.mode = state.mode
        self.inactive_commands = state.inactive_commands
        self.topstatus = state.topstatus

        self.list_args = state.list_args


STATE = State()
"""coBib's main-runtime State instance."""



"""Tests for coBib's TUI state."""
# pylint: disable=redefined-outer-name

import copy

import pytest

from cobib.config import config
from cobib.tui.state import STATE, Mode, State

from .. import get_resource


@pytest.fixture
def setup() -> None:
    """Setup."""
    config.load(get_resource("debug.py"))


def test_state_reset() -> None:
    """Test `cobib.tui.state.State.reset`."""
    state = copy.deepcopy(STATE)
    state.top_line = 10
    assert state.top_line == 10
    state.reset()
    assert state.top_line == 0


@pytest.mark.parametrize(
    ["reverse_order"],
    [
        [False],
        [True],
    ],
)
def test_state_initialize(reverse_order: bool) -> None:
    """Test config-dependent state-initialization.

    Args:
        reverse_order: whether to add the `--reverse` flag to the `default_list_args`.
    """
    config.tui.reverse_order = reverse_order
    state = copy.deepcopy(STATE)
    state.initialize()
    try:
        assert state.top_line == 0
        assert state.left_edge == 0
        assert state.current_line == 0
        assert state.previous_line == -1

        assert state.mode == Mode.LIST.value
        assert state.inactive_commands == []
        assert state.topstatus == ""

        state.list_args = config.tui.default_list_args
        if config.tui.reverse_order:
            state.list_args += ["-r"]
    finally:
        config.defaults()


def test_state_update() -> None:
    """Test `cobib.tui.state.State.update`."""
    state = copy.deepcopy(STATE)
    dummy_state = State()
    dummy_state.top_line = "dummy"  # type: ignore
    dummy_state.left_edge = "dummy"  # type: ignore
    dummy_state.current_line = "dummy"  # type: ignore
    dummy_state.previous_line = "dummy"  # type: ignore
    dummy_state.mode = "dummy"
    dummy_state.inactive_commands = "dummy"  # type: ignore
    dummy_state.topstatus = "dummy"
    dummy_state.list_args = "dummy"  # type: ignore

    state.update(dummy_state)
    assert dummy_state.top_line == "dummy"  # type: ignore
    assert dummy_state.left_edge == "dummy"  # type: ignore
    assert dummy_state.current_line == "dummy"  # type: ignore
    assert dummy_state.previous_line == "dummy"  # type: ignore
    assert dummy_state.mode == "dummy"
    assert dummy_state.inactive_commands == "dummy"  # type: ignore
    assert dummy_state.topstatus == "dummy"
    assert dummy_state.list_args == "dummy"  # type: ignore



"""coBib TUI test class."""

import curses
import fcntl
import logging
import os
import select
import signal
import struct
import tempfile
import termios
from datetime import date
from functools import partial
from pathlib import Path
from time import sleep
from typing import Any, Callable, Dict, List, Tuple, Union

import pyte
import pytest

from cobib.database import Database
from cobib.tui import TUI
from cobib.utils.logging import get_file_handler

from .mock_curses import MockCursesPad

TMP_LOGFILE = Path(tempfile.gettempdir()) / "cobib_tui_test.log"


class TUITest:
    """A TUI test runs coBib's TUI interface.

    The main part of this class is the `run_tui` method. The other methods are mainly setup and tear
    down utilities.
    """

    LOGGER = logging.getLogger("TUITest")
    """"We use a class-level LOGGER for more granular control."""

    @pytest.fixture
    def patch_curses(self, monkeypatch: pytest.MonkeyPatch) -> None:
        """Monkeypatch the `curses` module methods.

        Applies required changes to the `curses` module including set up of the `MockCursesPad`s.

        Args:
            monkeypatch: the built-in pytest fixture.
        """
        monkeypatch.setattr(
            "curses.curs_set", lambda *args: self.LOGGER.debug("curs_set: %s", args)
        )
        monkeypatch.setattr("curses.is_term_resized", lambda *args: True)
        monkeypatch.setattr("curses.resize_term", lambda *args: self.LOGGER.debug("resize_term"))
        monkeypatch.setattr(
            "curses.use_default_colors", lambda: self.LOGGER.debug("use_default_colors")
        )
        monkeypatch.setattr("curses.start_color", lambda: self.LOGGER.debug("start_color"))
        monkeypatch.setattr(
            "curses.init_pair", lambda *args: self.LOGGER.debug("init_pair: %s", args)
        )
        monkeypatch.setattr(
            "curses.init_color", lambda *args: self.LOGGER.debug("init_color: %s", args)
        )
        monkeypatch.setattr("curses.color_pair", lambda *args: args)
        monkeypatch.setattr("curses.pair_number", lambda *args: args)
        # pylint: disable=unnecessary-lambda
        monkeypatch.setattr("curses.pair_content", lambda *args: list(*args))
        monkeypatch.setattr("curses.newpad", lambda *args: MockCursesPad())
        monkeypatch.setattr("curses.newwin", lambda *args: MockCursesPad())
        monkeypatch.setattr("curses.endwin", lambda: self.LOGGER.debug("endwin"))

    @staticmethod
    def init_subprocess_coverage() -> Any:
        """Initializes the coverage reporting in a forked subprocess.

        This is required because the TUI itself will run in a forked subprocess. As such, we need to
        manually ensure that the coverage report gets updated accordingly.

        Returns:
            The coverage analyzer.
        """
        try:
            # pylint: disable=import-outside-toplevel
            import coverage
        except ImportError:
            return None
        _coveragerc = Path(__file__).parent.parent.parent / ".coveragerc"
        cov = coverage.Coverage(config_file=_coveragerc)
        cov.start()
        return cov

    @staticmethod
    def end_subprocess_coverage(*_, cov=None) -> None:  # type: ignore
        """Ends the subprocess coverage collection.

        Args:
            cov: the coverage analyzer to stop and save.
        """
        if cov is not None:
            cov.stop()
            cov.save()

    @staticmethod
    def run_tui(
        keys: Union[str, List[Union[str, signal.Signals]]],  # pylint: disable=no-member
        assertion: Callable,  # type: ignore
        assertion_kwargs: Dict,  # type: ignore
    ) -> None:
        """Spawns the coBib TUI in a forked pseudo-terminal.

        This method attaches a pyte object to the forked terminal to allow screen scraping. It also
        allows passing characters to the TUI by writing to the forked processes file descriptor.
        Furthermore, it also takes care of gathering the log, stdout/stderr and coverage information
        produced by the subprocess.

        For more information check out this
        [blog post](https://mrossinek.gitlab.io/programming/testing-tui-applications-in-python/).

        Args:
            keys: a string of characters passed to the TUI process ad verbatim.
            assertion: a callable method to assert the TUI state. This callable must take two
                       arguments: a pyte screen object and the caplog.record_tuples.
            assertion_kwargs: additional keyword arguments propagated to the assertion call.
        """
        # create pseudo-terminal
        pid, f_d = os.forkpty()

        if pid == 0:
            # setup subprocess coverage collection
            cov = TUITest.init_subprocess_coverage()
            signal.signal(signal.SIGTERM, partial(TUITest.end_subprocess_coverage, cov=cov))
            # redirect logging
            file_handler = get_file_handler(logging.DEBUG, TMP_LOGFILE)
            logging.getLogger().addHandler(file_handler)
            # child process initializes curses and spawns the TUI
            try:
                stdscr = curses.initscr()
                stdscr.resize(24, 80)
                if curses.has_colors():
                    curses.start_color()
                curses.cbreak()
                curses.noecho()
                stdscr.keypad(True)
                TUI(stdscr)
            finally:
                stdscr.keypad(False)
                curses.nocbreak()
                curses.echo()
        else:
            # parent process sets up virtual screen of identical size
            screen = pyte.Screen(80, 24)
            stream = pyte.ByteStream(screen)
            # send keys char-wise to TUI
            for key in keys:
                if key == signal.SIGWINCH:
                    sleep(0.25)
                    # resize pseudo terminal
                    buf = struct.pack("HHHH", 10, 45, 0, 0)
                    fcntl.ioctl(f_d, termios.TIOCSWINSZ, buf)
                    # overwrite screen
                    screen = pyte.Screen(45, 10)
                    stream = pyte.ByteStream(screen)
                else:
                    os.write(f_d, str.encode(str(key)))
            # scrape pseudo-terminal's screen
            while True:
                try:
                    [f_d], _, _ = select.select([f_d], [], [], 1)
                except (KeyboardInterrupt, ValueError):
                    # either test was interrupted or file descriptor of child process provides
                    # nothing to be read
                    break
                else:
                    try:
                        # scrape screen of child process
                        data = os.read(f_d, 1024)
                        stream.feed(data)
                    except OSError:
                        # reading empty
                        break
            # send SIGTERM to child process (which is necessary in order for pytest to not show
            # "failed" test cases produced by the forked child processes)
            os.kill(pid, signal.SIGTERM)
            # read the child process log file
            logs: List[Tuple[str, int, str]] = []
            print("### Captured logs from child process ###")
            with open(TMP_LOGFILE, "r", encoding="utf-8") as logfile:
                for line in logfile.readlines():
                    print(line.strip("\n"))
                    if not line.startswith(str(date.today())):
                        logs[-1] = (logs[-1][0], logs[-1][1], logs[-1][2] + line)
                        continue
                    split_line = line.split()
                    logs.append(
                        (
                            split_line[3],  # source
                            getattr(logging, split_line[2].strip("[]")),  # log level
                            " ".join(split_line[5:]),  # message
                        )
                    )
            os.remove(TMP_LOGFILE)
            # dump screen contents for easier debugging
            print("### Captured screen contents from child process ###")
            for line in screen.display:
                print(line)

            # It is necessary to read the database here, since we are in another subprocess
            Database().read()
            assertion(screen, logs, **assertion_kwargs)



"""coBib's TUI viewport."""

from __future__ import annotations

import copy
import curses
import logging
import re
from typing import TYPE_CHECKING, Dict, List, Optional, Tuple, Union

from cobib import __version__
from cobib.commands.list import ListCommand
from cobib.config import config

from .buffer import TextBuffer
from .state import STATE, Mode, State

LOGGER = logging.getLogger(__name__)

if TYPE_CHECKING:
    import cobib.tui


class Frame:
    """Frame class used to integrate a `TextBuffer` and `curses.pad` tightly with each other.

    This helper class is mainly used to implement the main 'viewport' of the TUI. It is a stateful
    object which handles a session-persistent history of the viewport to enable seamless level
    navigation.
    """

    def __init__(self, tui: cobib.tui.TUI, max_height: int, max_width: int) -> None:
        """Initializes the Frame object."""
        LOGGER.debug("Initializing frame's buffer")
        self.buffer = TextBuffer()
        LOGGER.debug("Initializing frame's pad")
        self.pad = curses.newpad(1, 1)
        # Also store a history of buffer contents and state (such as the current line, etc.)
        self.history: List[Tuple[TextBuffer, State]] = []

        # store TUI reference
        self.tui = tui
        self.height = max_height
        self.width = max_width

    def clear(self) -> None:
        """Wrapper for buffer.clear to intercept for history storage."""
        self.history.append((copy.deepcopy(self.buffer), copy.deepcopy(STATE)))
        # almost 100 entries should be much more than enough
        self.history = self.history[:99]
        self.buffer.clear()

    def revert(self) -> None:
        """Reverts the frame to the previous state."""
        if not self.history:
            LOGGER.debug("Empty frame history, nothing to revert")
            return
        self.buffer, state = self.history.pop()
        STATE.update(state)
        self.buffer.replace(
            range(self.buffer.height),
            re.escape(config.get_ansi_color("selection")) + r"(.+)" + re.escape("\x1b[0m"),
            r"\1",
        )
        # highlight current selection
        for label in self.tui.selection:
            # Note: this step may become a performance bottleneck because we replace inside the
            # whole buffer for each selected label!
            if STATE.mode == Mode.SEARCH.value:
                # Note: the inclusion of the search label is explained in the `SearchCommand`.
                self.buffer.replace(
                    range(self.buffer.height),
                    re.escape(config.get_ansi_color("search_label")) + label + re.escape("\x1b[0m"),
                    config.get_ansi_color("search_label")
                    + config.get_ansi_color("selection")
                    + label
                    + "\x1b[0m\x1b[0m",
                )
            elif STATE.mode == Mode.SHOW.value:
                self.buffer.replace(
                    0, label, config.get_ansi_color("selection") + label + "\x1b[0m"
                )
            else:
                # Note: the two spaces are explained in the `select()` method.
                self.buffer.replace(
                    range(self.buffer.height),
                    label + "  ",
                    config.get_ansi_color("selection") + label + "\x1b[0m  ",
                )
        self.view(ansi_map=self.tui.ANSI_MAP)
        self.tui.statusbar(self.tui.topbar, STATE.topstatus)

    def resize(self, new_height: int, new_width: int) -> None:
        """Resizes the Frame to the new dimensions.

        Args:
            new_height: the new height.
            new_width: the new width.
        """
        self.height = new_height
        self.width = new_width
        self.refresh()

    def refresh(self) -> None:
        """Utility function to quickly refresh the Frame's pad."""
        self.pad.refresh(STATE.top_line, STATE.left_edge, 1, 0, self.height, self.width - 1)

    def view(self, ansi_map: Optional[Dict[str, int]] = None) -> None:
        """Utility function to quickly view the Frame's buffer.

        Args:
            ansi_map: optional, dictionary mapping ANSI codes to curses color pairs.
        """
        self.buffer.view(self.pad, self.height, self.width - 1, ansi_map=ansi_map)

    def scroll_y(self, update: Union[int, str]) -> None:
        """Scroll vertically.

        Args:
            update: the offset specifying the scrolling height.
        """
        scrolloff = config.tui.scroll_offset
        overlap = scrolloff >= self.height - scrolloff
        scroll_lock = overlap and STATE.current_line - STATE.top_line == self.height // 2
        # jump to top
        if update == "g":
            LOGGER.debug("Jump to top of viewport.")
            STATE.top_line = 0
            STATE.current_line = 0
        # jump to bottom
        elif update == "G":
            LOGGER.debug("Jump to bottom of viewport.")
            STATE.top_line = max(self.buffer.height - self.height, 0)
            STATE.current_line = self.buffer.height - 1
        # scroll up
        elif isinstance(update, int) and update < 0:
            LOGGER.debug("Scroll viewport up by %d lines.", update)
            next_line = STATE.current_line + update
            if STATE.top_line > 0 and next_line < STATE.top_line + scrolloff:
                if scroll_lock or not overlap:
                    STATE.top_line += update
                elif (
                    overlap
                    and STATE.current_line - STATE.top_line > self.height // 2
                    and next_line - STATE.top_line < self.height // 2
                ):
                    STATE.top_line = next_line - self.height // 2
            STATE.top_line = max(STATE.top_line, 0)
            STATE.current_line = max(next_line, 0)
        # scroll down
        elif isinstance(update, int) and update > 0:
            LOGGER.debug("Scroll viewport down by %d lines.", update)
            next_line = STATE.current_line + update
            if (
                next_line >= STATE.top_line + self.height - scrolloff
                and self.buffer.height > STATE.top_line + self.height
            ):
                if scroll_lock or not overlap:
                    STATE.top_line += update
                elif (
                    overlap
                    and STATE.current_line - STATE.top_line < self.height // 2
                    and next_line - STATE.top_line > self.height // 2
                ):
                    STATE.top_line = next_line - self.height // 2
            if next_line < self.buffer.height:
                STATE.current_line = next_line
            else:
                STATE.top_line = self.buffer.height - self.height
                STATE.current_line = self.buffer.height - 1

    def scroll_x(self, update: Union[int, str]) -> None:
        """Scroll horizontally.

        Args:
            update: the offset specifying the scrolling width.
        """
        # jump to beginning
        if update == 0:
            LOGGER.debug("Jump to first column of viewport.")
            STATE.left_edge = 0
        # jump to end
        elif update == "$":
            LOGGER.debug("Jump to end of viewport.")
            STATE.left_edge = self.buffer.width - self.width
        elif isinstance(update, int):
            LOGGER.debug("Scroll viewport horizontally by %d columns.", update)
            next_col = STATE.left_edge + update
            # limit column such that no empty columns can appear on left or right
            if 0 <= next_col <= self.buffer.width - self.width:
                STATE.left_edge = next_col

    def wrap(self) -> None:
        """Toggles wrapping of the text currently displayed."""
        LOGGER.debug("Wrap command triggered.")
        # first, ensure left_edge is set to 0
        STATE.left_edge = 0
        # then, wrap the buffer
        self.buffer.wrap(self.width)
        self.view()
        # if cursor line is below buffer height, move it to the last line
        if self.buffer.height and STATE.current_line >= self.buffer.height:
            STATE.current_line = self.buffer.height - 1

    def get_current_label(self) -> Tuple[str, int]:
        """Returns the label and y position of the currently selected entry."""
        LOGGER.debug('Obtaining current label "under" cursor.')
        cur_y, _ = self.pad.getyx()
        # Two cases are possible: the list and the show mode
        if STATE.mode == Mode.LIST.value:
            # In the list mode, the label can be found in the current line
            # or in one of the previous lines if we are on a wrapped line
            while chr(self.pad.inch(cur_y, 0)) == TextBuffer.INDENT[0]:  # type: ignore
                cur_y -= 1
            label = self.pad.instr(cur_y, 0).decode("utf-8").split(" ")[0]  # type: ignore
        elif STATE.mode == Mode.SEARCH.value:
            # In the search mode, the same holds but we need to slightly change the label detection.
            while chr(self.pad.inch(cur_y, 0)) in ("[", TextBuffer.INDENT[0]):  # type: ignore
                cur_y -= 1
            label = self.pad.instr(cur_y, 0).decode("utf-8").split(" ")[0]  # type: ignore
        else:
            # In any other mode, the label can be found in the top statusbar
            label = "-".join(STATE.topstatus.split("-")[1:]).strip()
            # We also set cur_y to 0 for the select command to find it
            cur_y = 0
        label = re.sub(re.escape("\x1b[") + r".+m(.+)" + re.escape("\x1b[0m"), r"\1", label)
        LOGGER.debug('Current label at "%s" is "%s".', str(cur_y), label)
        return label, cur_y

    def update_list(self) -> None:
        """Updates the default list view."""
        LOGGER.debug("Re-populating the viewport with the list command.")
        self.buffer.clear()
        labels = ListCommand().execute(STATE.list_args, out=self.buffer)  # type: ignore
        labels = labels or []  # convert to empty list if labels is None
        # populate buffer with the list
        if STATE.mode != Mode.LIST.value:
            STATE.current_line = max(STATE.previous_line, 0)
            STATE.mode = Mode.LIST.value
        # reset viewport
        STATE.top_line = 0
        STATE.left_edge = 0
        STATE.inactive_commands = []
        # highlight current selection
        for label in self.tui.selection:
            # Note: the two spaces are explained in the `select()` method.
            # Also: this step may become a performance bottleneck because we replace inside the
            # whole buffer for each selected label!
            self.buffer.replace(
                range(self.buffer.height),
                label + "  ",
                config.get_ansi_color("selection") + label + "\x1b[0m  ",
            )
        # display buffer in viewport
        self.view(ansi_map=self.tui.ANSI_MAP)
        # update top statusbar
        STATE.topstatus = f"coBib v{__version__} - {len(labels)} Entries"
        self.tui.statusbar(self.tui.topbar, STATE.topstatus)
        # if cursor position is out-of-view (due to e.g. top-line reset in Show command), reset the
        # top-line such that the current line becomes height again
        if STATE.current_line > STATE.top_line + self.height:
            STATE.top_line = min(STATE.current_line, self.buffer.height - self.height)



"""Tests for coBib's DeleteCommand."""
# pylint: disable=no-self-use,unused-argument

from __future__ import annotations

import contextlib
import tempfile
from argparse import Namespace
from io import StringIO
from typing import TYPE_CHECKING, Any, List, Type

import pytest

from cobib.commands import DeleteCommand
from cobib.config import Event, config
from cobib.database import Database
from cobib.utils.rel_path import RelPath

from .. import get_resource
from ..tui.tui_test import TUITest
from .command_test import CommandTest

if TYPE_CHECKING:
    import cobib.commands


class TestDeleteCommand(CommandTest, TUITest):
    """Tests for coBib's DeleteCommand."""

    def get_command(self) -> Type[cobib.commands.base_command.Command]:
        # noqa: D102
        return DeleteCommand

    def _assert(self, labels: List[str]) -> None:
        """Common assertion utility method.

        Args:
            labels: the list of labels to be deleted.
        """
        bib = Database()

        for label in labels:
            assert bib.get(label, None) is None

        with open(config.database.file, "r", encoding="utf-8") as file:
            with open(get_resource("example_literature.yaml"), "r", encoding="utf-8") as expected:
                # NOTE: do NOT use zip_longest to omit last entries (for testing simplicity)
                for line, truth in zip(file, expected):
                    assert line == truth
                with pytest.raises(StopIteration):
                    file.__next__()

    @pytest.mark.parametrize(
        ["setup"],
        [
            [{"git": False}],
            [{"git": True}],
        ],
        indirect=["setup"],
    )
    @pytest.mark.parametrize(
        ["labels", "skip_commit"],
        [
            [["knuthwebsite"], False],
            [["knuthwebsite", "latexcompanion"], False],
            # non-existent labels should not cause any problems (but skip git check)
            [["dummy"], True],
            [["dummy", "knuthwebsite"], False],
        ],
    )
    def test_command(self, setup: Any, labels: List[str], skip_commit: bool) -> None:
        """Test the command itself.

        Args:
            setup: the `tests.commands.command_test.CommandTest.setup` fixture.
            labels: the list of labels to be deleted.
            skip_commit: whether to skip asserting the git commit message.
        """
        git = setup.get("git", False)

        # delete some data (for testing simplicity we delete the entries from the end)
        DeleteCommand().execute(labels)
        self._assert(labels)

        if git and not skip_commit:
            # assert the git commit message
            self.assert_git_commit_message("delete", {"labels": labels, "preserve_files": False})

    @pytest.mark.parametrize("preserve_files", [True, False])
    def test_remove_associated_file(self, setup: Any, preserve_files: bool) -> None:
        """Test removing associated files.

        Args:
            setup: the `tests.commands.command_test.CommandTest.setup` fixture.
            preserve_files: argument to `DeleteCommand`.
        """
        with tempfile.TemporaryDirectory() as tmpdirname:
            path = RelPath(tmpdirname + "/dummy.pdf")
            open(path.path, "w", encoding="utf-8").close()  # pylint: disable=consider-using-with

            Database()["knuthwebsite"].file = str(path)

            args = ["knuthwebsite"]
            if preserve_files:
                args += ["--preserve-files"]
            DeleteCommand().execute(args)

            assert path.path.exists() is preserve_files

    @pytest.mark.parametrize(["setup"], [[{"git": False}]], indirect=["setup"])
    def test_base_cmd_insufficient_git(self, setup: Any, caplog: pytest.LogCaptureFixture) -> None:
        """Test warning is raised by BaseCommand during insufficient git-configuration.

        While this is technically not related to the DeleteCommand, this is one of the faster
        commands to test this on.

        Args:
            setup: the `tests.commands.command_test.CommandTest.setup` fixture.
            caplog: the built-in pytest fixture.
        """
        config.database.git = True

        DeleteCommand().execute(["knuthwebsite"])
        self._assert(["knuthwebsite"])

        assert (
            "cobib.commands.base_command",
            30,
            "You have configured coBib to track your database with git."
            "\nPlease run `cobib init --git`, to initialize this tracking.",
        ) in caplog.record_tuples

    @pytest.mark.parametrize(
        ["labels"],
        [
            [["knuthwebsite"]],
            [["knuthwebsite", "latexcompanion"]],
        ],
    )
    # other variants are already covered by test_command
    def test_cmdline(self, setup: Any, monkeypatch: pytest.MonkeyPatch, labels: List[str]) -> None:
        """Test the command-line access of the command.

        Args:
            setup: the `tests.commands.command_test.CommandTest.setup` fixture.
            monkeypatch: the built-in pytest fixture.
            labels: the list of labels to be deleted.
        """
        self.run_module(monkeypatch, "main", ["cobib", "delete"] + labels)
        self._assert(labels)

    @pytest.mark.parametrize(
        ["select", "keys", "labels"],
        [
            [False, "vjvdq", ["knuthwebsite", "latexcompanion"]],
            [True, "d", ["knuthwebsite"]],
        ],
    )
    def test_tui(self, setup: Any, select: bool, keys: str, labels: List[str]) -> None:
        """Test the TUI access of the command.

        Args:
            setup: the `tests.commands.command_test.CommandTest.setup` fixture.
            select: whether to use the TUI selection.
            keys: the string of keys to pass to the TUI.
            labels: the list of labels to be deleted.
        """

        def assertion(screen, logs, **kwargs):  # type: ignore
            labels = kwargs.get("labels", [])
            self._assert(labels)

            true_log = [log for log in logs if log[0] == "cobib.commands.delete"]

            expected_log = [
                ("cobib.commands.delete", 10, "Delete command triggered from TUI."),
                ("cobib.commands.delete", 10, "Starting Delete command."),
            ]

            assert true_log[0:2] == expected_log

            # we cannot constructed a unique expected log because we do not know in which order the
            # labels are being removed (because the list of labels gets converted from an unordered
            # set)
            for label in labels:
                assert (
                    "cobib.commands.delete",
                    10,
                    f"Attempting to delete entry '{label}'.",
                ) in true_log
                assert (
                    "cobib.commands.delete",
                    20,
                    f"'{label}' was removed from the database.",
                ) in true_log

                # also assert that the label is no longer visible on screen
                assert all(label not in line for line in screen.display[1:-3])

            assert true_log[-1] == (
                "cobib.commands.delete",
                10,
                "Updating list after Delete command.",
            )

        self.run_tui(keys, assertion, {"labels": labels})

    def test_event_pre_delete_command(self, setup: Any) -> None:
        """Tests the PreDeleteCommand event."""

        @Event.PreDeleteCommand.subscribe
        def hook(largs: Namespace) -> None:
            largs.labels = ["einstein"]

        assert Event.PreDeleteCommand.validate()

        DeleteCommand().execute(["knuthwebsite"])

        assert "einstein" not in Database().keys()
        assert "knuthwebsite" in Database().keys()

    def test_event_post_delete_command(self, setup: Any) -> None:
        """Tests the PostDeleteCommand event."""

        @Event.PostDeleteCommand.subscribe
        def hook(deleted_entries: List[str]) -> None:
            for label in deleted_entries:
                print(f"WARNING: deleted entry '{label}'")

        with contextlib.redirect_stdout(StringIO()) as out:
            DeleteCommand().execute(["knuthwebsite"])

        assert Event.PostDeleteCommand.validate()

        assert out.getvalue().strip() == "WARNING: deleted entry 'knuthwebsite'"



"""Tests for coBib's SearchCommand."""
# pylint: disable=no-self-use,unused-argument

from __future__ import annotations

import contextlib
import re
from argparse import Namespace
from io import StringIO
from itertools import zip_longest
from typing import TYPE_CHECKING, Any, List, Type

import pytest

from cobib.commands import SearchCommand
from cobib.config import Event, config

from ..tui.tui_test import TUITest
from .command_test import CommandTest

if TYPE_CHECKING:
    import cobib.commands


class TestSearchCommand(CommandTest, TUITest):
    """Tests for coBib's SearchCommand."""

    def get_command(self) -> Type[cobib.commands.base_command.Command]:
        # noqa: D102
        return SearchCommand

    def _assert(self, output: List[str], expected: List[str]) -> None:
        """Common assertion utility method.

        Args:
            output: the list of lines printed to `sys.stdout`.
            expected: the expected output.
        """
        # we use zip_longest to ensure that we don't have more than we expect
        for line, exp in zip_longest(output, expected):
            line = line.replace("\x1b", "")
            line = re.sub(r"\[[0-9;]+m", "", line)
            if exp:
                assert exp in line
            if line and not (line.endswith("match") or line.endswith("matches")):
                assert re.match(r"\[[0-9]+\]", line)

    @pytest.mark.parametrize(
        ["args", "expected", "config_overwrite"],
        [
            [
                ["einstein"],
                ["einstein - 1 match", "@article{einstein,", "author = {Albert Einstein},"],
                False,
            ],
            [
                ["einstein", "-i"],
                [
                    "einstein - 2 matches",
                    "@article{einstein,",
                    "author = {Albert Einstein},",
                    "doi = {http://dx.doi.org/10.1002/andp.19053221004},",
                ],
                False,
            ],
            [
                ["einstein", "-i", "-c", "0"],
                ["einstein - 2 matches", "@article{einstein,", "author = {Albert Einstein},"],
                False,
            ],
            [
                ["einstein", "-i", "-c", "2"],
                [
                    "einstein - 2 matches",
                    "@article{einstein,",
                    "author = {Albert Einstein},",
                    "doi = {http://dx.doi.org/10.1002/andp.19053221004},",
                    "journal = {Annalen der Physik},",
                ],
                False,
            ],
            [
                ["einstein"],
                [
                    "einstein - 2 matches",
                    "@article{einstein,",
                    "author = {Albert Einstein},",
                    "doi = {http://dx.doi.org/10.1002/andp.19053221004},",
                ],
                True,
            ],
            [
                ["einstein", "-i"],
                [
                    "einstein - 2 matches",
                    "@article{einstein,",
                    "author = {Albert Einstein},",
                    "doi = {http://dx.doi.org/10.1002/andp.19053221004},",
                ],
                True,
            ],
        ],
    )
    def test_command(
        self, setup: Any, args: List[str], expected: List[str], config_overwrite: bool
    ) -> None:
        """Test the command itself.

        Args:
            setup: the `tests.commands.command_test.CommandTest.setup` fixture.
            args: the arguments to pass to the command.
            expected: the expected output.
            config_overwrite: what to overwrite `config.commands.search.ignore_case` with.
        """
        config.commands.search.ignore_case = config_overwrite
        file = StringIO()

        SearchCommand().execute(args, out=file)
        self._assert(file.getvalue().split("\n"), expected)

    @pytest.mark.parametrize(
        ["expected"],
        [
            [["einstein - 1 match", "@article{einstein,", "author = {Albert Einstein},"]],
        ],
    )
    # other variants are already covered by test_command
    def test_cmdline(
        self,
        setup: Any,
        monkeypatch: pytest.MonkeyPatch,
        capsys: pytest.CaptureFixture[str],
        expected: List[str],
    ) -> None:
        """Test the command-line access of the command.

        Args:
            setup: the `tests.commands.command_test.CommandTest.setup` fixture.
            monkeypatch: the built-in pytest fixture.
            capsys: the built-in pytest fixture.
            expected: the expected output.
        """
        self.run_module(monkeypatch, "main", ["cobib", "search", "einstein"])
        self._assert(capsys.readouterr().out.strip().split("\n"), expected)

    @pytest.mark.parametrize(
        ["select", "keys"],
        [
            [False, "/einstein\n"],
            [True, "Gv/einstein\n"],
        ],
    )
    def test_tui(self, setup: Any, select: bool, keys: str) -> None:
        """Test the TUI access of the command.

        Args:
            setup: the `tests.commands.command_test.CommandTest.setup` fixture.
            select: whether to use the TUI selection.
            keys: the string of keys to pass to the TUI.
        """

        def assertion(screen, logs, **kwargs):  # type: ignore
            expected_screen = [
                "einstein - 1 match",
                "[1]     @article{einstein,",
                "[1]      author = {Albert Einstein},",
            ]
            for line, truth in zip(expected_screen, screen.display[1:]):
                assert line == truth.strip()

            expected_log = [
                ("cobib.commands.search", 10, "Search command triggered from TUI."),
                ("cobib.commands.search", 10, "Starting Search command."),
                (
                    "cobib.commands.search",
                    10,
                    "Available entries to search: ['einstein', 'latexcompanion', 'knuthwebsite']",
                ),
                ("cobib.commands.search", 10, "The search will be performed case sensitive"),
                ("cobib.commands.search", 10, 'Entry "einstein" includes 1 hits.'),
                ("cobib.commands.search", 10, "Applying selection highlighting in search results."),
                ("cobib.commands.search", 10, "Populating viewport with search results."),
                ("cobib.commands.search", 10, "Resetting cursor position to top."),
            ]
            assert [log for log in logs if log[0] == "cobib.commands.search"] == expected_log

            colors = []
            for _ in range(3):
                # color matrix is initialized with default values for three lines
                colors.append([("default", "default")] * 80)

            # selection must be visible
            if select:
                colors[0][0:8] = [("white", "magenta")] * 8
            else:
                colors[0][0:8] = [("blue", "black")] * 8
            # the current line highlighting is only visible from here onward
            colors[0][8:] = [("white", "cyan")] * 72
            # the search keyword highlighting
            colors[1][17:25] = [("red", "black")] * 8
            for idx, color_line in enumerate(colors):
                assert [(c.fg, c.bg) for c in screen.buffer[idx + 1].values()] == color_line

        self.run_tui(keys, assertion, {"selection": select})

    def test_tui_no_hits(self, setup: Any) -> None:
        """Test how the TUI handles no search results.

        Args:
            setup: the `tests.commands.command_test.CommandTest.setup` fixture.
        """

        def assertion(screen, logs, **kwargs):  # type: ignore
            expected_screen = [
                "knuthwebsite    Knuth: Computers and Typesetting",
                r"latexcompanion  The \LaTeX\ Companion",
                r"einstein        Zur Elektrodynamik bewegter K{\"o}rper",
            ]
            for line, truth in zip(expected_screen, screen.display[1:]):
                assert line == truth.strip()

            assert "No search hits for 'dummy'!" in screen.display[-1]

            expected_log = [
                ("cobib.commands.search", 10, "Search command triggered from TUI."),
                ("cobib.commands.search", 10, "Starting Search command."),
                (
                    "cobib.commands.search",
                    10,
                    "Available entries to search: ['einstein', 'latexcompanion', 'knuthwebsite']",
                ),
                ("cobib.commands.search", 10, "The search will be performed case sensitive"),
                ("cobib.commands.search", 20, "No search hits for 'dummy'!"),
            ]
            assert [log for log in logs if log[0] == "cobib.commands.search"] == expected_log

        self.run_tui("/dummy\n", assertion, {})

    def test_event_pre_search_command(self, setup: Any) -> None:
        """Tests the PreSearchCommand event."""

        @Event.PreSearchCommand.subscribe
        def hook(largs: Namespace) -> None:
            largs.query = "einstein"

        assert Event.PreSearchCommand.validate()

        expected = ["einstein - 1 match", "@article{einstein,", "author = {Albert Einstein},"]

        file = StringIO()
        SearchCommand().execute(["knuthwebsite"], out=file)
        self._assert(file.getvalue().split("\n"), expected)

    def test_event_post_search_command(self, setup: Any) -> None:
        """Tests the PostSearchCommand event."""

        @Event.PostSearchCommand.subscribe
        def hook(hits: int, labels: List[str]) -> None:
            print(labels)

        assert Event.PostSearchCommand.validate()

        with contextlib.redirect_stdout(StringIO()) as file:
            SearchCommand().execute(["einstein"])
            out = file.getvalue().split("\n")
            self._assert(out[:-2], [])
            assert out[-2] == "['einstein']"



"""coBib's Show command.

This command simply shows/prints the specified entry as a BibLaTex-formmatted string.
```
cobib show <label>
```

You can also trigger this command from the `cobib.tui.tui.TUI`.
By default, it is bound to the `ENTER` key.
"""

from __future__ import annotations

import argparse
import logging
import sys
from typing import IO, TYPE_CHECKING, Any, List

from cobib import __version__
from cobib.config import Event, config
from cobib.database import Database
from cobib.parsers.bibtex import BibtexParser

from .base_command import ArgumentParser, Command

LOGGER = logging.getLogger(__name__)

if TYPE_CHECKING:
    import cobib.tui


class ShowCommand(Command):
    """The Show Command."""

    name = "show"

    def execute(self, args: List[str], out: IO[Any] = sys.stdout) -> None:
        """Shows an entry in its BibLaTex-format.

        This command simply shows/prints the specified entry as a BibLaTex-formmatted string.

        Args:
            args: a sequence of additional arguments used for the execution. The following values
                are allowed for this command:
                    * `label`: the label label of the entry to be shown.
            out: the output IO stream. This defaults to `sys.stdout`.
        """
        LOGGER.debug("Starting Show command.")
        parser = ArgumentParser(prog="show", description="Show subcommand parser.")
        parser.add_argument("label", type=str, help="label of the entry")

        if not args:
            parser.print_usage(sys.stderr)
            sys.exit(1)

        try:
            largs = parser.parse_args(args)
        except argparse.ArgumentError as exc:
            LOGGER.error(exc.message)
            return

        Event.PreShowCommand.fire(largs)

        try:
            entry = Database()[largs.label]
            entry_str = BibtexParser().dump(entry)

            entry_str = Event.PostShowCommand.fire(entry_str) or entry_str

            print(entry_str, file=out)
        except KeyError:
            msg = f"No entry with the label '{largs.label}' could be found."
            LOGGER.error(msg)

    @staticmethod
    def tui(tui: cobib.tui.TUI) -> None:
        # pdoc will inherit the docstring from the base class
        # noqa: D102
        LOGGER.debug("Show command triggered from TUI.")
        # get current label
        label, cur_y = tui.viewport.get_current_label()
        # populate buffer with entry data
        LOGGER.debug("Clearing current buffer contents.")
        tui.viewport.clear()
        ShowCommand().execute([label], out=tui.viewport.buffer)  # type: ignore
        tui.viewport.buffer.split()
        if label in tui.selection:
            LOGGER.debug("Current entry is selected. Applying highlighting.")
            tui.viewport.buffer.replace(
                0, label, config.get_ansi_color("selection") + label + "\x1b[0m"
            )
        LOGGER.debug("Populating buffer with ShowCommand result.")
        tui.viewport.view(ansi_map=tui.ANSI_MAP)

        # reset current cursor position
        tui.STATE.top_line = 0
        tui.STATE.current_line = 0
        # update top statusbar
        tui.STATE.topstatus = f"coBib v{__version__} - {label}"
        tui.statusbar(tui.topbar, tui.STATE.topstatus)
        # enter show menu
        tui.STATE.mode = "show"
        tui.STATE.previous_line = cur_y
        tui.STATE.inactive_commands = ["Add", "Filter", "Show", "Sort"]



"""Tests for coBib's TUI Buffer."""

from typing import Any, Dict, List, Optional, Union

import pytest

from cobib.config import config
from cobib.tui import TUI, InputBuffer, TextBuffer

from .mock_curses import MockCursesPad
from .mock_tui import MockTUI


class TestTextBuffer:
    """Tests for coBib's TextBuffer."""

    @pytest.fixture(autouse=True)
    def setup(self) -> None:
        """Setup.

        This fixture is automatically enabled for all tests in this class.
        """
        # pylint: disable=attribute-defined-outside-init
        self.buffer = TextBuffer()

    @pytest.mark.parametrize(
        ["strings", "expected"],
        [
            [[""], {"lines": [], "height": 0, "width": 0, "wrapped": False, "ansi_map": None}],
            [
                ["test"],
                {"lines": ["test"], "height": 1, "width": 4, "wrapped": False, "ansi_map": None},
            ],
            [
                ["test\ntest"],
                {
                    "lines": ["test\ntest"],
                    "height": 1,
                    "width": 9,
                    "wrapped": False,
                    "ansi_map": None,
                },
            ],
            [
                ["test", "test"],
                {
                    "lines": ["test", "test"],
                    "height": 2,
                    "width": 4,
                    "wrapped": False,
                    "ansi_map": None,
                },
            ],
        ],
    )
    def test_write(self, strings: List[str], expected: Dict[str, Any]) -> None:
        """Test `cobib.tui.buffer.TextBuffer.write`.

        Args:
            strings: a list of strings to place into the buffer.
            expected: a dictionary with the expected values for the buffer's internal state.
        """
        for string in strings:
            self.buffer.write(string)
        for key, value in expected.items():
            assert getattr(self.buffer, key) == value

    @pytest.mark.parametrize(
        ["lines", "old", "new", "expected"],
        [
            [0, "test", "tmp", ["tmp0", "test1", "test2"]],
            [[0, 1], "test", "tmp", ["tmp0", "tmp1", "test2"]],
            [[0, 2], "test", "tmp", ["tmp0", "test1", "tmp2"]],
            [[0, 1, 2], "test", "tmp", ["tmp0", "tmp1", "tmp2"]],
        ],
    )
    def test_replace(
        self, lines: Union[int, List[int]], old: str, new: str, expected: List[str]
    ) -> None:
        """Test `cobib.tui.buffer.TextBuffer.replace`.

        Args:
            lines: the line numbers on which to replace.
            old: the string to be replaced.
            new: the string to be inserted.
            expected: the expected resulting list of strings in the buffer.
        """
        for string in ["test" + str(num) for num in range(3)]:
            self.buffer.write(string)
        self.buffer.replace(lines, old, new)
        assert self.buffer.lines == expected

    def test_clear(self) -> None:
        """Test `cobib.tui.buffer.TextBuffer.clear`."""
        self.buffer.lines = ["test"]
        self.buffer.clear()
        assert self.buffer.lines == []
        assert self.buffer.height == 0
        assert self.buffer.width == 0
        assert self.buffer.wrapped is False

    def test_split(self) -> None:
        """Test `cobib.tui.buffer.TextBuffer.split`."""
        self.buffer.lines = ["test\ntest"]
        self.buffer.split()
        assert self.buffer.lines == ["test", "test"]
        assert self.buffer.height == 2
        assert self.buffer.width == 4

    @pytest.mark.parametrize(
        ["width", "label_column", "expected"],
        [
            [
                20,
                False,
                [
                    "Label0  Title0 by",
                    "↪  Author0",
                    "Label1  Title1 by",
                    "↪  Author1",
                    "Label2  Title2 by",
                    "↪  Author2",
                    "Label3  Title3 by",
                    "↪  Author3",
                    "Label4  Title4 by",
                    "↪  Author4",
                ],
            ],
            [
                20,
                True,
                [
                    "Label0  Title0 by",
                    "↪       Author0",
                    "Label1  Title1 by",
                    "↪       Author1",
                    "Label2  Title2 by",
                    "↪       Author2",
                    "Label3  Title3 by",
                    "↪       Author3",
                    "Label4  Title4 by",
                    "↪       Author4",
                ],
            ],
            [
                6,
                True,
                [
                    "Label",
                    "↪ 0",
                    "↪ Tit",
                    "↪ le0",
                    "↪ by",
                    "↪ Aut",
                    "↪ hor",
                    "↪ 0",
                    "Label",
                    "↪ 1",
                    "↪ Tit",
                    "↪ le1",
                    "↪ by",
                    "↪ Aut",
                    "↪ hor",
                    "↪ 1",
                    "Label",
                    "↪ 2",
                    "↪ Tit",
                    "↪ le2",
                    "↪ by",
                    "↪ Aut",
                    "↪ hor",
                    "↪ 2",
                    "Label",
                    "↪ 3",
                    "↪ Tit",
                    "↪ le3",
                    "↪ by",
                    "↪ Aut",
                    "↪ hor",
                    "↪ 3",
                    "Label",
                    "↪ 4",
                    "↪ Tit",
                    "↪ le4",
                    "↪ by",
                    "↪ Aut",
                    "↪ hor",
                    "↪ 4",
                ],
            ],
        ],
    )
    def test_wrap(self, width: int, label_column: bool, expected: List[str]) -> None:
        """Test `cobib.tui.buffer.TextBuffer.wrap`.

        Args:
            width: the number of columns at which to wrap the buffer.
            label_column: whether to determine a label column width or not.
            expected: the expected resulting list of strings in the buffer.
        """
        self.buffer.lines = [
            "Label0  Title0 by Author0",
            "Label1  Title1 by Author1",
            "Label2  Title2 by Author2",
            "Label3  Title3 by Author3",
            "Label4  Title4 by Author4",
        ]
        self.buffer.wrap(width, label_column)
        for line, truth in zip(self.buffer.lines, expected):
            assert line.strip() == truth

    def test_unwrap(self) -> None:
        """Test `cobib.tui.buffer.TextBuffer.unwrap`."""
        self.buffer.wrapped = True
        self.buffer.lines = [
            "Label0  Title0 by",
            "↪  Author0",
            "Label1  Title1 by",
            "↪  Author1",
            "Label2  Title2 by",
            "↪  Author2",
            "Label3  Title3 by",
            "↪  Author3",
            "Label4  Title4 by",
            "↪  Author4",
        ]
        self.buffer.wrap(10)
        assert self.buffer.lines == [
            "Label0  Title0 by Author0",
            "Label1  Title1 by Author1",
            "Label2  Title2 by Author2",
            "Label3  Title3 by Author3",
            "Label4  Title4 by Author4",
        ]

    def test_view(self, caplog: pytest.LogCaptureFixture) -> None:
        """Test `cobib.tui.buffer.TextBuffer.view`.

        Args:
            caplog: the built-in pytest fixture.
        """
        self.buffer.lines = [
            "Label0  Title0 by Author0",
            "Label1  Title1 by Author1",
            "Label2  Title2 by Author2",
            "Label3  Title3 by Author3",
            "Label4  Title4 by Author4",
        ]
        pad = MockCursesPad()
        self.buffer.view(pad, 10, 40)
        expected_log = [
            ("MockCursesPad", 10, "erase"),
            ("MockCursesPad", 10, "refresh: 0 0 1 0 10 40"),
            ("MockCursesPad", 10, "resize: 1 41"),
            ("MockCursesPad", 10, "addstr: 0 0 Label0  Title0 by Author0"),
            ("MockCursesPad", 10, "addstr: 1 0 Label1  Title1 by Author1"),
            ("MockCursesPad", 10, "addstr: 2 0 Label2  Title2 by Author2"),
            ("MockCursesPad", 10, "addstr: 3 0 Label3  Title3 by Author3"),
            ("MockCursesPad", 10, "addstr: 4 0 Label4  Title4 by Author4"),
            ("MockCursesPad", 10, "refresh: 0 0 1 0 10 40"),
        ]
        assert [
            record for record in caplog.record_tuples if record[0] == "MockCursesPad"
        ] == expected_log

    def test_view_with_box(self, caplog: pytest.LogCaptureFixture) -> None:
        """Test `cobib.tui.buffer.TextBuffer.view` with a surrounding box.

        Args:
            caplog: the built-in pytest fixture.
        """
        self.buffer.lines = [
            "Label0  Title0 by Author0",
            "Label1  Title1 by Author1",
            "Label2  Title2 by Author2",
            "Label3  Title3 by Author3",
            "Label4  Title4 by Author4",
        ]
        pad = MockCursesPad()
        self.buffer.view(pad, 10, 40, box=True)
        expected_log = [
            ("MockCursesPad", 10, "erase"),
            ("MockCursesPad", 10, "refresh: 0 0 1 0 10 40"),
            ("MockCursesPad", 10, "resize: 2 40"),
            ("MockCursesPad", 10, "addstr: 1 1 Label0  Title0 by Author0"),
            ("MockCursesPad", 10, "addstr: 2 1 Label1  Title1 by Author1"),
            ("MockCursesPad", 10, "addstr: 3 1 Label2  Title2 by Author2"),
            ("MockCursesPad", 10, "addstr: 4 1 Label3  Title3 by Author3"),
            ("MockCursesPad", 10, "addstr: 5 1 Label4  Title4 by Author4"),
            ("MockCursesPad", 10, "box"),
            ("MockCursesPad", 10, "refresh: 0 0 1 0 10 40"),
        ]
        assert [
            record for record in caplog.record_tuples if record[0] == "MockCursesPad"
        ] == expected_log

    def test_view_with_ansi_map(
        self, caplog: pytest.LogCaptureFixture, monkeypatch: pytest.MonkeyPatch
    ) -> None:
        """Test `cobib.tui.buffer.TextBuffer.view` with an ANSI color map.

        Args:
            caplog: the built-in pytest fixture.
            monkeypatch: the built-in pytest fixture.
        """
        monkeypatch.setattr("curses.color_pair", lambda *args: args)
        # create ANSI color map for testing purposes
        config.defaults()
        ansi_map = {}
        for attr in TUI.COLOR_NAMES:
            ansi_map[config.get_ansi_color(attr)] = TUI.COLOR_NAMES.index(attr) + 1
        # populate buffer
        self.buffer.lines = [
            "\x1b[31;40mLabel0\x1b[0m  Title0 by Author0",
            "\x1b[31;40mLabel1\x1b[0m  Title1 by Author1",
            "\x1b[31;40mLabel2\x1b[0m  Title2 by Author2",
            "\x1b[31;40mLabel3\x1b[0m  Title3 by Author3",
            "\x1b[31;40mLabel4\x1b[0m  Title4 by Author4",
        ]
        pad = MockCursesPad()
        self.buffer.view(pad, 10, 40, ansi_map=ansi_map)
        expected_log = [
            ("MockCursesPad", 10, "erase"),
            ("MockCursesPad", 10, "refresh: 0 0 1 0 10 40"),
            ("MockCursesPad", 10, "resize: 1 41"),
            ("MockCursesPad", 10, "addstr: 0 0 Label0  Title0 by Author0"),
            ("MockCursesPad", 10, "chgat: 0 0 6 (4,)"),
            ("MockCursesPad", 10, "addstr: 1 0 Label1  Title1 by Author1"),
            ("MockCursesPad", 10, "chgat: 1 0 6 (4,)"),
            ("MockCursesPad", 10, "addstr: 2 0 Label2  Title2 by Author2"),
            ("MockCursesPad", 10, "chgat: 2 0 6 (4,)"),
            ("MockCursesPad", 10, "addstr: 3 0 Label3  Title3 by Author3"),
            ("MockCursesPad", 10, "chgat: 3 0 6 (4,)"),
            ("MockCursesPad", 10, "addstr: 4 0 Label4  Title4 by Author4"),
            ("MockCursesPad", 10, "chgat: 4 0 6 (4,)"),
            ("MockCursesPad", 10, "refresh: 0 0 1 0 10 40"),
        ]
        assert [
            record for record in caplog.record_tuples if record[0] == "MockCursesPad"
        ] == expected_log
        for message in ["Interpreting ANSI color codes on the fly.", "Applying ANSI color map."]:
            assert ("cobib.tui.buffer", 10, message) in caplog.record_tuples

    def test_view_with_bkgd(
        self, caplog: pytest.LogCaptureFixture, monkeypatch: pytest.MonkeyPatch
    ) -> None:
        """Test `cobib.tui.buffer.TextBuffer.view` with a background color.

        Args:
            caplog: the built-in pytest fixture.
            monkeypatch: the built-in pytest fixture.
        """
        monkeypatch.setattr("curses.color_pair", lambda *args: args)
        self.buffer.lines = [
            "Label0  Title0 by Author0",
            "Label1  Title1 by Author1",
            "Label2  Title2 by Author2",
            "Label3  Title3 by Author3",
            "Label4  Title4 by Author4",
        ]
        pad = MockCursesPad()
        self.buffer.view(pad, 10, 40, background=1)
        expected_log = [
            ("MockCursesPad", 10, "erase"),
            ("MockCursesPad", 10, "refresh: 0 0 1 0 10 40"),
            ("MockCursesPad", 10, "resize: 1 41"),
            ("MockCursesPad", 10, "addstr: 0 0 Label0  Title0 by Author0"),
            ("MockCursesPad", 10, "addstr: 1 0 Label1  Title1 by Author1"),
            ("MockCursesPad", 10, "addstr: 2 0 Label2  Title2 by Author2"),
            ("MockCursesPad", 10, "addstr: 3 0 Label3  Title3 by Author3"),
            ("MockCursesPad", 10, "addstr: 4 0 Label4  Title4 by Author4"),
            ("MockCursesPad", 10, "bkgd:   (2,)"),
            ("MockCursesPad", 10, "refresh: 0 0 1 0 10 40"),
        ]
        assert [
            record for record in caplog.record_tuples if record[0] == "MockCursesPad"
        ] == expected_log

    def test_view_with_bkgd_and_box(
        self, caplog: pytest.LogCaptureFixture, monkeypatch: pytest.MonkeyPatch
    ) -> None:
        """Test `cobib.tui.buffer.TextBuffer.view` with a background color and surrounding box.

        Args:
            caplog: the built-in pytest fixture.
            monkeypatch: the built-in pytest fixture.
        """
        monkeypatch.setattr("curses.color_pair", lambda *args: args)
        self.buffer.lines = [
            "Label0  Title0 by Author0",
            "Label1  Title1 by Author1",
            "Label2  Title2 by Author2",
            "Label3  Title3 by Author3",
            "Label4  Title4 by Author4",
        ]
        pad = MockCursesPad()
        self.buffer.view(pad, 10, 40, background=1, box=True)
        expected_log = [
            ("MockCursesPad", 10, "erase"),
            ("MockCursesPad", 10, "refresh: 0 0 1 0 10 40"),
            ("MockCursesPad", 10, "resize: 2 40"),
            ("MockCursesPad", 10, "addstr: 1 1 Label0  Title0 by Author0"),
            ("MockCursesPad", 10, "addstr: 2 1 Label1  Title1 by Author1"),
            ("MockCursesPad", 10, "addstr: 3 1 Label2  Title2 by Author2"),
            ("MockCursesPad", 10, "addstr: 4 1 Label3  Title3 by Author3"),
            ("MockCursesPad", 10, "addstr: 5 1 Label4  Title4 by Author4"),
            ("MockCursesPad", 10, "bkgd:   (2,)"),
            ("MockCursesPad", 10, "box"),
            ("MockCursesPad", 10, "refresh: 0 0 1 0 10 40"),
        ]
        assert [
            record for record in caplog.record_tuples if record[0] == "MockCursesPad"
        ] == expected_log

    @pytest.mark.parametrize(
        ["background"],
        [
            [None],
            [1],
        ],
    )
    def test_popup(
        self,
        caplog: pytest.LogCaptureFixture,
        monkeypatch: pytest.MonkeyPatch,
        background: Optional[int],
    ) -> None:
        """Test `cobib.tui.buffer.TextBuffer.popup`.

        Args:
            caplog: the built-in pytest fixture.
            monkeypatch: the built-in pytest fixture.
            background: the value of the background color.
        """
        monkeypatch.setattr("curses.newpad", lambda *args: MockCursesPad())
        if background is not None:
            monkeypatch.setattr("curses.color_pair", lambda *args: args)
        self.buffer.lines = [
            "Label0  Title0 by Author0",
            "Label1  Title1 by Author1",
            "Label2  Title2 by Author2",
            "Label3  Title3 by Author3",
            "Label4  Title4 by Author4",
        ]
        self.buffer.popup(MockTUI(), background=background)  # type: ignore
        expected_log = [
            ("MockCursesPad", 10, "erase"),
            ("MockCursesPad", 10, "refresh: 0 0 16 0 19 40"),
            ("MockCursesPad", 10, "resize: 7 40"),
            ("MockCursesPad", 10, "addstr: 1 1 Label0  Title0 by Author0"),
            ("MockCursesPad", 10, "addstr: 2 1 Label1  Title1 by Author1"),
            ("MockCursesPad", 10, "addstr: 3 1 Label2  Title2 by Author2"),
            ("MockCursesPad", 10, "addstr: 4 1 Label3  Title3 by Author3"),
            ("MockCursesPad", 10, "addstr: 5 1 Label4  Title4 by Author4"),
            ("MockCursesPad", 10, "box"),
            ("MockCursesPad", 10, "refresh: 0 0 16 0 19 40"),
            ("MockCursesPad", 10, "getch"),
            ("MockCursesPad", 10, "clear"),
            ("MockTUI", 10, "resize_handler"),
        ]
        if background is not None:
            expected_log.insert(8, ("MockCursesPad", 10, "bkgd:   (2,)"))
        assert [
            record for record in caplog.record_tuples if record[0] in ("MockCursesPad", "MockTUI")
        ] == expected_log


def test_input_buffer(caplog: pytest.LogCaptureFixture, monkeypatch: pytest.MonkeyPatch) -> None:
    """Test the `cobib.tui.buffer.InputBuffer`.

    Since this buffer is a very minimal wrapper, this simple test suffices.

    Args:
        caplog: the built-in pytest fixture.
        monkeypatch: the built-in pytest fixture.
    """
    monkeypatch.setattr("curses.newpad", lambda *args: MockCursesPad())
    buffer = InputBuffer(TextBuffer(), MockTUI())  # type: ignore
    buffer.readline()
    expected_log = [
        ("MockCursesPad", 10, "erase"),
        ("MockCursesPad", 10, "refresh: 0 0 16 0 18 40"),
        ("MockCursesPad", 10, "resize: 2 40"),
        ("MockCursesPad", 10, "box"),
        ("MockCursesPad", 10, "refresh: 0 0 16 0 18 40"),
        ("MockTUI", 10, "prompt_handler"),
        ("MockCursesPad", 10, "clear"),
        ("MockTUI", 10, "resize_handler"),
    ]
    assert [
        record for record in caplog.record_tuples if record[0] in ("MockCursesPad", "MockTUI")
    ] == expected_log



"""Mocking utilities for a `curses.window`."""
# pylint: disable=missing-function-docstring

import logging


class MockCursesPad:
    """This class mocks a `curses.window`.

    When testing the `cobib.tui` modules, we need full control over the `curses` objects. Thus,
    this class removes all of the non-essential aspects and logs all method calls.
    """

    def __init__(self, lines=None) -> None:  # type: ignore
        # noqa: D107
        self.logger = logging.getLogger("MockCursesPad")
        self.lines = lines or []
        self.current_pos = [0, 0]
        self.size = (0, 0)
        self.returned_chars = [27]  # Escape

    def erase(self):  # type: ignore
        # noqa: D102
        self.logger.debug("erase")

    def refresh(  # type: ignore
        self, pminrow=None, pmincol=None, sminrow=None, smincol=None, smaxrow=None, smaxcol=None
    ):
        # noqa: D102
        self.logger.debug(
            "refresh: %s %s %s %s %s %s", pminrow, pmincol, sminrow, smincol, smaxrow, smaxcol
        )

    def resize(self, nlines, ncols):  # type: ignore
        # noqa: D102
        self.logger.debug("resize: %s %s", nlines, ncols)
        self.size = (nlines, ncols)

    def mvwin(self, new_y, new_x):  # type: ignore
        # noqa: D102
        self.logger.debug("mvwin: %s %s", new_y, new_x)

    def insstr(self, row, col, string):  # type: ignore
        # noqa: D102
        self.logger.debug("insstr: %s %s %s", row, col, string)
        try:
            self.lines[row] = string
        except IndexError:
            self.lines.insert(row, string)

    def addstr(self, row, col, string):  # type: ignore
        # noqa: D102
        self.logger.debug("addstr: %s %s %s", row, col, string)
        try:
            old_string = self.lines[row]
        except IndexError:
            while len(self.lines) <= row:
                self.lines.append("")
            old_string = ""
        new_string = old_string[:col] + string + old_string[col + len(string) :]
        self.lines[row] = new_string
        print(self.lines)

    def addnstr(self, row, col, string, num, attr):  # type: ignore
        # noqa: D102
        self.logger.debug("addnstr: %s %s %s %s %s", row, col, string, num, attr)
        try:
            old_string = self.lines[row]
        except IndexError:
            while len(self.lines) <= row:
                self.lines.append([])
            old_string = ""
        new_string = old_string[:col] + string + old_string[col + len(string) :]
        self.lines[row] = new_string

    def chgat(self, row, col, num, attr):  # type: ignore
        # noqa: D102
        self.logger.debug("chgat: %s %s %s %s", row, col, num, attr)

    def bkgd(self, char, attr):  # type: ignore
        # noqa: D102
        self.logger.debug("bkgd: %s %s", char, attr)

    def box(self):  # type: ignore
        # noqa: D102
        self.logger.debug("box")

    def clear(self):  # type: ignore
        # noqa: D102
        self.logger.debug("clear")

    def getch(self):  # type: ignore
        # noqa: D102
        self.logger.debug("getch")
        return self.returned_chars.pop()

    def inch(self, cur_y, cur_x):  # type: ignore
        # noqa: D102
        self.logger.debug("inch: %s, %s", cur_y, cur_x)
        try:
            return ord(self.lines[cur_y][cur_x])
        except IndexError:
            return -1

    def instr(self, cur_y, cur_x):  # type: ignore
        # noqa: D102
        self.logger.debug("instr: %s, %s", cur_y, cur_x)
        return self.lines[cur_y][cur_x:].encode()

    def getyx(self):  # type: ignore
        # noqa: D102
        self.logger.debug("getyx")
        return tuple(self.current_pos)

    def getmaxyx(self):  # type: ignore
        # noqa: D102
        self.logger.debug("getmaxyx")
        return self.size

    def move(self, new_y, new_x):  # type: ignore
        # noqa: D102
        self.logger.debug("move: %s, %s", new_y, new_x)
        self.current_pos = [new_y, new_x]

    def nodelay(self, flag):  # type: ignore
        # noqa: D102
        self.logger.debug("nodelay: %s", flag)

    def delch(self, row, col):  # type: ignore
        # noqa: D102
        self.logger.debug("delch: %s, %s", row, col)
        line = self.lines[row]
        self.lines[row] = line[:col] + line[col + 1 :]
        self.current_pos[1] -= 1



"""coBib's Init command.

This command must be the first command you ever execute with coBib because it takes care of
initializing the database.
Generally, you will only ever need to run this command once, but it is safe to run it multiple times
(although it will likely have no effect).

To get started with coBib you must run:
```
cobib init
```
This will initialize the database in the location specified by `config.database.file`.

If you enabled the automatic git-integration of coBib via `config.database.git`, you must initialize
this separately via:
```
cobib init --git
```
If you have not run the first command yet, you can directly initialize the database *and* the
git-integration by only running the second command.
"""

from __future__ import annotations

import argparse
import logging
import os
import sys
from typing import IO, TYPE_CHECKING, Any, List

from cobib.config import Event, config
from cobib.utils.rel_path import RelPath

from .base_command import ArgumentParser, Command

LOGGER = logging.getLogger(__name__)

if TYPE_CHECKING:
    import cobib.tui


class InitCommand(Command):
    """The Init Command."""

    name = "init"

    def execute(self, args: List[str], out: IO[Any] = sys.stdout) -> None:
        """Initializes the database.

        Initializes the YAML database in the location specified by `config.database.file`.
        If you enabled `config.database.git` *and* you specify the `--git` command-line argument,
        the git-integration will be initialized, too.

        Args:
            args: a sequence of additional arguments used for the execution. The following values
                are allowed for this command:
                    * `-g`, `--git`: initializes the git-integration.
            out: the output IO stream. This defaults to `sys.stdout`.
        """
        LOGGER.debug("Starting Init command.")
        parser = ArgumentParser(prog="init", description="Init subcommand parser.")
        parser.add_argument("-g", "--git", action="store_true", help="initialize git repository")

        try:
            largs = parser.parse_args(args)
        except argparse.ArgumentError as exc:
            LOGGER.error(exc.message)
            return

        Event.PreInitCommand.fire(largs)

        file = RelPath(config.database.file).path
        root = file.parent

        file_exists = file.exists()
        git_tracked = (root / ".git").exists()

        if file_exists:
            if git_tracked:
                msg = (
                    "Database file already exists and is being tracked by git. There is nothing "
                    "else to do."
                )
                LOGGER.info(msg)
                return

            if not git_tracked and not largs.git:
                msg = "Database file already exists! Use --git to start tracking it with git."
                LOGGER.warning(msg)
                return

        else:
            LOGGER.debug('Creating path for database file: "%s"', root)
            root.mkdir(parents=True, exist_ok=True)

            LOGGER.debug('Creating empty database file: "%s"', file)
            open(file, "w", encoding="utf-8").close()  # pylint: disable=consider-using-with

        if largs.git:
            if not config.database.git:
                msg = (
                    "You are about to initialize the git tracking of your database, but this will "
                    "only have effect if you also enable the DATABASE/git setting in your "
                    "configuration file!"
                )
                LOGGER.warning(msg)
            # First, check whether git is configured correctly.
            print("Checking `git config --get user.name`:", end=" ", flush=True)
            name_set = os.system("git config --get user.name")
            print()
            print("Checking `git config --get user.email`:", end=" ", flush=True)
            email_set = os.system("git config --get user.email")
            print()
            if name_set != 0 or email_set != 0:
                msg = (
                    "In order to use git you must configure your name and email first! For more "
                    "information please consult `man gittutorial`."
                )
                LOGGER.warning(msg)
                sys.exit(1)
            LOGGER.debug('Initializing git repository in "%s"', root)
            os.system(f"git init {root}")
            self.git(args=vars(largs), force=True)

        Event.PostInitCommand.fire(root, file)

    @staticmethod
    def tui(tui: cobib.tui.TUI) -> None:
        """🛑 This command is *not* available via the TUI.

        This is the case because the TUI cannot be started before the database has been initialized.
        """



"""Tests for coBib's config validation."""
# pylint: disable=unused-argument, redefined-outer-name

import logging
import tempfile
from typing import Any, Generator, List

import pytest

from cobib.config import config
from cobib.config.config import Config

from .. import get_resource

EXAMPLE_LITERATURE = get_resource("example_literature.yaml")


def test_config_init() -> None:
    """Test Config initialization."""
    # empty init
    assert Config() == {}
    # init from empty dictionary
    assert Config({}) == {}
    # init from dictionary
    assert Config({"dummy": "test"}) == {"dummy": "test"}
    # init from non-dictionary
    with pytest.raises(TypeError):
        Config(True)  # type: ignore
    with pytest.raises(TypeError):
        Config(1)  # type: ignore
    with pytest.raises(TypeError):
        Config("")  # type: ignore
    with pytest.raises(TypeError):
        Config([])  # type: ignore


def test_config_getattr() -> None:
    """Test the automatic attribute generation."""
    cfg = Config()
    dummy = cfg.dummy
    assert isinstance(cfg.dummy, Config)
    assert dummy == {}


def test_config_recursive_getattr() -> None:
    """Test the recursive attribute generation."""
    cfg = Config()
    dummy = cfg.dummy.dummy
    assert isinstance(cfg.dummy, Config)
    assert isinstance(cfg.dummy.dummy, Config)
    assert dummy == {}


def test_config_recursive_setattr() -> None:
    """Test the recursive attribute setting."""
    cfg = Config()
    cfg.dummy.dummy = "test"
    assert isinstance(cfg.dummy, Config)
    assert cfg.dummy.dummy == "test"
    assert cfg.dummy == {"dummy": "test"}


def test_config_load() -> None:
    """Test loading another config file."""
    config.load(get_resource("debug.py"))
    assert config.database.file == str(EXAMPLE_LITERATURE)


def test_config_load_from_open_file() -> None:
    """Test loading another config from an open file."""
    with open(get_resource("debug.py"), "r", encoding="utf-8") as file:
        config.load(file)
    assert config.database.file == str(EXAMPLE_LITERATURE)


def test_config_load_nothing() -> None:
    """Test that nothing changes when no XDG files are present."""
    Config.XDG_CONFIG_FILE = ""
    Config.LEGACY_XDG_CONFIG_FILE = ""
    config.load()
    # we manually call validate because load exits early
    config.validate()


def test_config_load_xdg() -> None:
    """Test loading a config from XDG path."""
    Config.XDG_CONFIG_FILE = get_resource("debug.py")
    config.load()
    assert config.database.file == str(EXAMPLE_LITERATURE)


# TODO: remove legacy configuration support on 1.1.2022
def assert_legacy_config() -> None:
    """Assert the legacy configuration has been applied."""
    assert config.commands.edit.default_entry_type == "string"
    assert config.commands.open.command == "string"
    assert config.commands.search.grep == "string"
    assert config.commands.search.ignore_case is True
    assert config.database.file == "string"
    assert config.database.git is True
    assert config.parsers.bibtex.ignore_non_standard_types is True
    assert config.tui.default_list_args == ["string"]
    assert config.tui.prompt_before_quit is False
    assert config.tui.reverse_order is False
    assert config.tui.scroll_offset == 5
    assert config.tui.colors.cursor_line_fg == "black"
    assert config.tui.key_bindings.prompt == "p"


def test_config_load_legacy() -> None:
    """Test loading a legacy config file."""
    config.load_legacy_config(get_resource("legacy_config.ini", "config"))
    # first, it must pass the validation test
    config.validate()
    # then we also check that all settings have been changed somehow
    assert_legacy_config()


def test_config_load_legacy_xdg() -> None:
    """Test loading a legacy config from XDG path."""
    Config.XDG_CONFIG_FILE = ""
    Config.LEGACY_XDG_CONFIG_FILE = get_resource("legacy_config.ini", "config")
    config.load()  # validation is done internally
    # then we also check that all settings have been changed somehow
    assert_legacy_config()


def test_config_example() -> None:
    """Test that the example config matches the default values."""
    config.clear()
    config.load(get_resource("example.py", "../src/cobib/config/"))
    assert config == Config.DEFAULTS


def test_config_validation_failure(caplog: pytest.LogCaptureFixture) -> None:
    """Test for a `SystemExit` upon config validation failure.

    Args:
        caplog: the built-in pytest fixture.
    """
    with pytest.raises(SystemExit):
        config.load(get_resource("broken_config.py", "config"))
    assert (
        "cobib.config.config",
        logging.ERROR,
        "config.database.file should be a string.",
    ) in caplog.record_tuples


@pytest.fixture
def setup() -> Generator[Any, None, None]:
    """Setup debugging configuration.

    Yields:
        Access to the local fixture variables.
    """
    config.load(get_resource("debug.py"))
    yield setup
    config.clear()
    config.defaults()


def test_config_validation(setup: Any) -> None:
    """Test that the initial configuration passes all validation checks.

    Args:
        setup: a local pytest fixture.
    """
    config.validate()


@pytest.mark.parametrize(
    ["sections", "field"],
    [
        [["logging"], "logfile"],
        [["logging"], "version"],
        [["commands", "edit"], "default_entry_type"],
        [["commands", "edit"], "editor"],
        [["commands", "open"], "command"],
        [["commands", "search"], "grep"],
        [["commands", "search"], "grep_args"],
        [["commands", "search"], "ignore_case"],
        [["database"], "file"],
        [["database"], "git"],
        [["database", "format"], "label_default"],
        [["database", "format"], "label_suffix"],
        [["database", "format"], "suppress_latex_warnings"],
        [["database", "stringify", "list_separator"], "file"],
        [["database", "stringify", "list_separator"], "tags"],
        [["database", "stringify", "list_separator"], "url"],
        [["parsers", "bibtex"], "ignore_non_standard_types"],
        [["tui"], "default_list_args"],
        [["tui"], "prompt_before_quit"],
        [["tui"], "reverse_order"],
        [["tui"], "scroll_offset"],
        [["tui", "colors"], "cursor_line_fg"],
        [["tui", "colors"], "cursor_line_bg"],
        [["tui", "colors"], "top_statusbar_fg"],
        [["tui", "colors"], "top_statusbar_bg"],
        [["tui", "colors"], "bottom_statusbar_fg"],
        [["tui", "colors"], "bottom_statusbar_bg"],
        [["tui", "colors"], "search_label_fg"],
        [["tui", "colors"], "search_label_bg"],
        [["tui", "colors"], "search_query_fg"],
        [["tui", "colors"], "search_query_bg"],
        [["tui", "colors"], "popup_help_fg"],
        [["tui", "colors"], "popup_help_bg"],
        [["tui", "colors"], "popup_stdout_fg"],
        [["tui", "colors"], "popup_stdout_bg"],
        [["tui", "colors"], "popup_stderr_fg"],
        [["tui", "colors"], "popup_stderr_bg"],
        [["tui", "colors"], "selection_fg"],
        [["tui", "colors"], "selection_bg"],
        [["tui", "key_bindings"], "prompt"],
        [["tui", "key_bindings"], "search"],
        [["tui", "key_bindings"], "help"],
        [["tui", "key_bindings"], "add"],
        [["tui", "key_bindings"], "delete"],
        [["tui", "key_bindings"], "edit"],
        [["tui", "key_bindings"], "filter"],
        [["tui", "key_bindings"], "modify"],
        [["tui", "key_bindings"], "open"],
        [["tui", "key_bindings"], "quit"],
        [["tui", "key_bindings"], "redo"],
        [["tui", "key_bindings"], "sort"],
        [["tui", "key_bindings"], "undo"],
        [["tui", "key_bindings"], "select"],
        [["tui", "key_bindings"], "wrap"],
        [["tui", "key_bindings"], "export"],
        [["tui", "key_bindings"], "show"],
        [["utils", "file_downloader"], "default_location"],
        [["utils", "file_downloader"], "url_map"],
        [["utils"], "journal_abbreviations"],
    ],
)
def test_missing_config_fields(setup: Any, sections: List[str], field: str) -> None:
    """Test raised RuntimeError for missing config fields.

    Args:
        setup: a local pytest fixture.
        sections: a list of section names in the nested configuration.
        field: the name of the configuration setting.
    """
    with pytest.raises(RuntimeError) as exc_info:
        section = config
        for sec in sections[:-1]:
            section = section[sec]
        del section[sections[-1]][field]
        config.validate()
    assert f"config.{'.'.join(sections)}.{field}" in str(exc_info.value)


@pytest.mark.parametrize(
    ["color"],
    [
        ["cursor_line_fg"],
        ["cursor_line_bg"],
        ["top_statusbar_fg"],
        ["top_statusbar_bg"],
        ["bottom_statusbar_fg"],
        ["bottom_statusbar_bg"],
        ["search_label_fg"],
        ["search_label_bg"],
        ["search_query_fg"],
        ["search_query_bg"],
        ["popup_help_fg"],
        ["popup_help_bg"],
        ["popup_stdout_fg"],
        ["popup_stdout_bg"],
        ["popup_stderr_fg"],
        ["popup_stderr_bg"],
        ["selection_fg"],
        ["selection_bg"],
    ],
)
def test_valid_tui_colors(setup: Any, color: str) -> None:
    """Test curses color specification validation.

    Args:
        setup: a local pytest fixture.
        color: the name of the color.
    """
    with pytest.raises(RuntimeError) as exc_info:
        config.tui.colors[color] = "test"
        config.validate()
    assert str(exc_info.value) == "Unknown color specification: test"


@pytest.mark.parametrize(
    ["color", "ansi"],
    [
        ["cursor_line", "\x1b[37;46m"],
        ["top_statusbar", "\x1b[30;43m"],
        ["bottom_statusbar", "\x1b[30;43m"],
        ["search_label", "\x1b[34;40m"],
        ["search_query", "\x1b[31;40m"],
        ["popup_help", "\x1b[37;42m"],
        ["popup_stdout", "\x1b[37;44m"],
        ["popup_stderr", "\x1b[37;41m"],
        ["selection", "\x1b[37;45m"],
    ],
)
def test_get_ansi_color(setup: Any, color: str, ansi: str) -> None:
    """Test default ANSI color code generation.

    Args:
        setup: a local pytest fixture.
        color: the name of the color.
        ansi: the expected ANSI code.
    """
    assert config.get_ansi_color(color) == ansi


def test_ignored_tui_color(setup: Any, caplog: pytest.LogCaptureFixture) -> None:
    """Test invalid TUI colors are ignored.

    Args:
        setup: a local pytest fixture.
        caplog: the built-in pytest fixture.
    """
    config.tui.colors.dummy = "white"
    config.validate()
    assert (
        "cobib.config.config",
        logging.WARNING,
        "Ignoring unknown TUI color: dummy.",
    ) in caplog.record_tuples


@pytest.mark.parametrize(["setting", "value"], [[["database", "format", "month"], str]])
def test_deprecation_warning(
    setting: List[str], value: Any, caplog: pytest.LogCaptureFixture
) -> None:
    """Test logged warning for deprecated setting.

    Args:
        setting: the list of attribute names leading to the deprecated setting.
        value: a value to use for the deprecated setting.
        caplog: the built-in pytest fixture.
    """
    section = config
    for sec in setting[:-1]:
        section = section[sec]
    section[setting[-1]] = value
    config.validate()
    for source, level, message in caplog.record_tuples:
        if (
            source == "cobib.config.config"
            and level == logging.WARNING
            and f"The config.{'.'.join(setting)} setting is deprecated" in message
        ):
            break
    else:
        assert False, "Missing deprecation warning!"


@pytest.mark.parametrize(
    ["setting", "attribute"], [["[FORMAT]\nmonth=str", "database.format.month"]]
)
def test_deprecation_warning_legacy(
    setting: str, attribute: str, caplog: pytest.LogCaptureFixture
) -> None:
    """Test logged warning for deprecated setting.

    Args:
        setting: the legacy formatted string of the deprecated setting.
        attribute: the new formatted name of the deprecated setting.
        caplog: the built-in pytest fixture.
    """
    with tempfile.NamedTemporaryFile("w") as legacy_file:
        legacy_file.write(setting)
        legacy_file.seek(0)
        config.load_legacy_config(legacy_file.name)
        config.validate()
    for source, level, message in caplog.record_tuples:
        if (
            source == "cobib.config.config"
            and level == logging.WARNING
            and f"The config.{attribute} setting is deprecated" in message
        ):
            break
    else:
        assert False, "Missing deprecation warning!"



r"""coBib's Modify command.

This command allows you to perform bulk modification to multiple entries.
Thus, it provides faster means to apply simple edits to many entries at once, without having to open
each entry for editing one-by-one or having to edit the database file manually.

A simple example is the following:
```
cobib modify tags:private --selection -- Label1 Label2 ...
```
which will set the tags of all listed entries to `private`.

You can use the `--add` option to not overwrite but append to existing values of a field. `str`
fields will be concatenated with*out* any spaces, lists will be appended, and numeric fields will be
added. Any other kind of field will be converted to a `str`.

As with other commands, you can also use filters (see also `cobib.commands.list`) rather than a
manual selection to specify the entries which to modify:
```
cobib modify tags:first_author -- ++author Rossmannek
```

As of v3.2.0 the `value` provided as part of the modification is interpreted as an "f"-string.[^1]
This means you can even use placeholder variables and perform simple operations on them. The
available variables depend on the entry which you are modifying as they are inferred from its stored
data.
Below are some examples:
```
# Rewrite the 'pages' field with a single-dash separator
cobib modify "pages:{pages.replace('--', '-')}" -- ...

# Rename an entry according to the first author's surname and year
cobib modify "label:{author.split()[1]}{year}" -- ...
```

In combination with the regex-support for filters added during the same release, you can even unify
your database's label convention:
```
cobib modify "label:{label.replace('_', '')}" -- ++label "\D+_\d+"
```

If you happen to use an undefined variable as part of your modification, coBib will handle this
gracefully by falling back to an empty string and raising a warning:
```
cobib modify "note:{undefined}" -- ...
> [WARNING] You tried use an undefined variable. Falling back to an empty string.
> [ERROR] name 'undefined' is not defined
```

Since v3.3.0 this command also provides a "dry" mode which previews the modifications without
actually applying them.
```
cobib modify --dry <modification> -- ...
```
This is useful if you want to test large bulk modifications before running them in order to prevent
mistakes.

You can also trigger this command from the `cobib.tui.tui.TUI`.
By default, it is bound to the `m` key which will drop you into the prompt where you can type out a
normal command-line command:
```
:modify <arguments go here>
```

[^1]: <https://docs.python.org/3/reference/lexical_analysis.html#formatted-string-literals>
"""

from __future__ import annotations

import argparse
import ast
import logging
import os
import sys
from typing import IO, TYPE_CHECKING, Any, Callable, Dict, List, Optional, Tuple

from cobib.config import Event
from cobib.database import Database
from cobib.utils.logging import get_stream_handler
from cobib.utils.rel_path import RelPath

from .base_command import ArgumentParser, Command
from .list import ListCommand

LOGGER = logging.getLogger(__name__)

if TYPE_CHECKING:
    import cobib.tui


def evaluate_ast_node(node: ast.expr, locals_: Optional[Dict[str, Any]] = None) -> str:
    """Evaluates an AST node representing an f-string.

    Args:
        node: the AST expression extracted from an f-string.
        locals_: the dictionary of local variables to be used as context for the expression
            evaluation.

    Returns:
        The evaluated AST node expression.
    """
    try:
        # pylint: disable=eval-used
        return eval(  # type: ignore
            compile(ast.Expression(node), filename="<string>", mode="eval"), locals_
        )
    except NameError as err:
        LOGGER.warning("You tried use an undefined variable. Falling back to an empty string.")
        LOGGER.error(err)
        return ""


def evaluate_as_f_string(value: str, locals_: Optional[Dict[str, Any]] = None) -> str:
    """Evaluates a string as if it were a literal f-string.

    Args:
        value: the string to be evaluated.
        locals_: the dictionary of local variables to be used as context for the expression
            evaluation.

    Returns:
        The evaluated f-string.

    Raises:
        ValueError: if an unexpected AST component type is encountered.

    References:
        <https://stackoverflow.com/a/61190684>
    """
    result: List[str] = []
    for part in ast.parse(f"f'''{value}'''").body[0].value.values:  # type: ignore
        typ = type(part)

        if typ is ast.Constant:
            result.append(part.value)

        elif typ is ast.Str:
            # TODO: remove once support for Python 3.7 will be dropped
            result.append(part.s)

        elif typ is ast.FormattedValue:
            value = evaluate_ast_node(part.value, locals_)

            if part.conversion >= 0:
                conversions: Dict[str, Callable[[Any], str]] = {"a": ascii, "r": repr, "s": str}
                value = conversions[chr(part.conversion)](value)

            if part.format_spec:
                value = format(value, evaluate_ast_node(part.format_spec))

            result.append(str(value))

        else:
            LOGGER.warning("Unexpected AST node expression type '%s' for an f-string.", typ)
            raise ValueError

    return "".join(result)


class ModifyCommand(Command):
    """The Modify Command."""

    name = "modify"

    @staticmethod
    def field_value_pair(string: str) -> Tuple[str, str]:
        """Utility method to assert the field-value pair argument type.

        This method is given to the `argparse.ArgumentParser` instance as its `type` specifier.
        An input argument is considered valid if it passes through this function without raising any
        errors, which means it conforms to the `<field>:<value>` notation.

        Args:
            string: the argument string to check.
        """
        # try splitting the string into field and value, any errors will be handled by argparse
        field, *value = string.split(":")
        # NOTE: we split only the first field off in case the value contains f-string format
        # specifications
        return (field, ":".join(value))

    # pylint: disable=too-many-branches,too-many-statements
    def execute(self, args: List[str], out: IO[Any] = sys.stdout) -> None:
        """Modifies multiple entries in bulk.

        This command allows bulk modification of multiple entries.
        It takes a modification in the form `<field>:<value>` and will overwrite the `field` of all
        matching entries with the new `value`.
        The entries can be specified as a manual selection (when using `--selection` or the visual
        selection of the TUI) or through filters (see also `cobib.commands.list`).

        Args:
            args: a sequence of additional arguments used for the execution. The following values
                are allowed for this command:
                    * `modification`: a string conforming to `<field>:<value>` indicating the
                      modification that should be applied to all matching entries. By default, the
                      modification will overwrite any existing data in the specified `field` with
                      the new `value`. For more information about formatting options of `<value>`
                      refer to the module documentation or the man-page.
                    * `--dry`: run in "dry"-mode which lists modifications without applying them.
                    * `-a`, `--add`: when specified, the modification's value will be added to the
                      entry's field rather than overwrite it. If the field in question is numeric,
                      the numbers will be added.
                    * `-s`, `--selection`: when specified, the positional arguments will *not* be
                      interpreted as filters but rather as a direct list of entry labels. This can
                      be used on the command-line but is mainly meant for the TUIs visual selection
                      interface (hence the name).
                    * in addition to the above, you can add `filters` to specify a subset of your
                      database for exporting. For more information refer to `cobib.commands.list`.
            out: the output IO stream. This defaults to `sys.stdout`.
        """
        LOGGER.debug("Starting Modify command.")
        parser = ArgumentParser(prog="modify", description="Modify subcommand parser.")
        parser.add_argument(
            "modification",
            type=self.field_value_pair,
            help="Modification to apply to the specified entries."
            "\nThis argument must be a string formatted as <field>:<value> where field can be any "
            "field of the entries and value can be any string which should be placed in that "
            "field. Be sure to escape this field-value pair properly, especially if the value "
            "contains spaces.",
        )
        parser.add_argument(
            "--dry",
            action="store_true",
            help="Run in 'dry'-mode, listing modifications without actually applying them.",
        )
        parser.add_argument(
            "-a",
            "--add",
            action="store_true",
            help="Adds to the modified field rather than overwriting it.",
        )
        parser.add_argument(
            "-s",
            "--selection",
            action="store_true",
            help="When specified, the `filter` argument will be interpreted as a list of entry "
            "labels rather than arguments for the `list` command.",
        )
        parser.add_argument(
            "filter",
            nargs="+",
            help="You can specify filters as used by the `list` command in order to select a "
            "subset of labels to be modified. To ensure this works as expected you should add the "
            "pseudo-argument '--' before the list of filters. See also `list --help` for more "
            "information.",
        )
        parser.add_argument(
            "--preserve-files", action="store_true", help="do not rename associated files"
        )

        if not args:
            parser.print_usage(sys.stderr)
            sys.exit(1)

        try:
            largs = parser.parse_intermixed_args(args)
        except argparse.ArgumentError as exc:
            LOGGER.error(exc.message)
            return

        Event.PreModifyCommand.fire(largs)

        info_handler: logging.Handler
        if largs.dry:
            info_handler = get_stream_handler(logging.INFO)

            class ModifyInfoFilter(logging.Filter):
                """A logging filter to only print ModifyCommand INFO messages."""

                def filter(self, record: logging.LogRecord) -> bool:
                    return record.name == "cobib.commands.modify" and record.levelname == "INFO"

            info_handler.addFilter(ModifyInfoFilter())
            LOGGER.addHandler(info_handler)

        if largs.selection:
            LOGGER.info("Selection given. Interpreting `filter` as a list of labels")
            labels = largs.filter
        else:
            LOGGER.debug("Gathering filtered list of entries to be modified.")
            with open(os.devnull, "w", encoding="utf-8") as devnull:
                labels = ListCommand().execute(largs.filter, out=devnull)

        field, value = largs.modification

        bib = Database()

        for label in labels:  # pylint: disable=too-many-nested-blocks
            try:
                entry = bib[label]
                local_value = evaluate_as_f_string(value, {"label": label, **entry.data.copy()})

                if hasattr(entry, field):
                    prev_value = getattr(entry, field, None)
                else:
                    prev_value = entry.data.get(field, None)

                if not largs.add:
                    new_value = local_value
                    if local_value.isnumeric():
                        new_value = int(local_value)  # type: ignore
                else:
                    try:
                        if prev_value is None:
                            new_value = local_value
                        elif isinstance(prev_value, str):
                            new_value = prev_value + local_value
                        elif isinstance(prev_value, list):
                            new_value = prev_value + [local_value]  # type: ignore
                        elif isinstance(prev_value, int):
                            if local_value.isnumeric():
                                new_value = prev_value + int(local_value)  # type: ignore
                            else:
                                raise TypeError
                        else:
                            raise TypeError
                    except TypeError:
                        LOGGER.warning(
                            "Encountered an unexpected field type to add to. Converting the field "
                            "'%s' of entry '%s' to a simple string: '%s'.",
                            field,
                            label,
                            str(prev_value) + local_value,
                        )
                        new_value = str(prev_value) + local_value

                # guard against overwriting existing data if label gets changed
                if field == "label":
                    new_value = bib.disambiguate_label(new_value)

                if new_value == prev_value:
                    LOGGER.info(
                        "New and previous values match. Skipping modification of entry '%s'.", label
                    )
                    continue

                if hasattr(entry, field):
                    if largs.dry:
                        LOGGER.info(
                            "%s: changing field '%s' from %s to %s",
                            entry.label,
                            field,
                            getattr(entry, field),
                            new_value,
                        )
                    setattr(entry, field, new_value)
                else:
                    if largs.dry:
                        LOGGER.info(
                            "%s: adding field '%s' = %s",
                            entry.label,
                            field,
                            new_value,
                        )
                    entry.data[field] = new_value

                bib.update({entry.label: entry})

                if entry.label != label:
                    bib.rename(label, entry.label)
                    if not largs.preserve_files:
                        new_files = []
                        for file in entry.file:
                            path = RelPath(file)
                            if path.path.stem == label:
                                LOGGER.info("Also renaming associated file '%s'.", str(path))
                                target = RelPath(path.path.parent / f"{entry.label}.pdf")
                                if target.path.exists():
                                    LOGGER.warning(
                                        "Found conflicting file, not renaming '%s'.", str(path)
                                    )
                                else:
                                    if largs.dry:
                                        LOGGER.info(
                                            "%s: renaming associated file '%s' to '%s'",
                                            entry.label,
                                            path.path,
                                            target.path,
                                        )
                                    else:
                                        path.path.rename(target.path)
                                        new_files.append(str(target))
                                    continue
                            if not largs.dry:
                                new_files.append(file)
                        if not largs.dry:
                            entry.file = new_files

                if not largs.dry:
                    msg = f"'{label}' was modified."
                    LOGGER.info(msg)
            except KeyError:
                msg = f"No entry with the label '{label}' could be found."
                LOGGER.warning(msg)

        Event.PostModifyCommand.fire(labels, largs.dry)

        if largs.dry:
            LOGGER.removeHandler(info_handler)
            # read also functions as a restoring method
            bib.read()
        else:
            bib.save()
            self.git(args=vars(largs))

    @staticmethod
    def tui(tui: cobib.tui.TUI) -> None:
        # pdoc will inherit the docstring from the base class
        # noqa: D102
        LOGGER.debug("Modify command triggered from TUI.")
        # handle input via prompt
        if tui.selection:
            tui.execute_command("modify -s", pass_selection=True)
        else:
            tui.execute_command("modify")



"""coBib's Search command.

This command allows you to search your database for a regex-interpreted query.
While doing so, it uses the `config.commands.search.grep` tool to search associated files, too.

As a simple example, you can query for a simple author name like so:
```
cobib search Einstein
```
You can make the search case *in*sensitive in two ways:
1. By enabling `config.commands.search.ignore_case`.
2. By providing the `--ignore-case` command-line argument:
```
cobib search --ignore-case Einstein
```

By default, the search command will provide you with 1 line of context above and below the actual
matches. You can change this number of lines by setting the `--context` option:
```
cobib search --context 4 Einstein
```

Finally, you can also combine the search with coBib's filtering mechanism to narrow your search down
to a subset of your database:
```
cobib search Einstein -- ++year 2020
```
Note, that we use the auxiliary `--` argument to separate the filters from the actual arguments.
While this is not strictly necessary it helps to disambiguate the origin of the arguments.

You can also trigger this command from the `cobib.tui.tui.TUI`.
By default, it is bound to the `/` key.
"""

from __future__ import annotations

import argparse
import logging
import os
import re
import shlex
import sys
from typing import IO, TYPE_CHECKING, Any, List, Optional, Tuple

from cobib import __version__
from cobib.config import Event, config
from cobib.database import Database

from .base_command import ArgumentParser, Command
from .list import ListCommand

LOGGER = logging.getLogger(__name__)

if TYPE_CHECKING:
    import cobib.tui


class SearchCommand(Command):
    """The Search Command."""

    name = "search"

    def execute(
        self, args: List[str], out: Optional[IO[Any]] = None
    ) -> Optional[Tuple[int, List[str]]]:
        """Searches in the database.

        This command searches the database for a regex-interpreted query.
        It leverages `cobib.database.Entry.search` to perform the actual search.

        You can configure the search-tool which searches through associated files via
        `config.commands.search.grep`.

        Args:
            args: a sequence of additional arguments used for the execution. The following values
                are allowed for this command:
                    * `query`: the required positional argument corresponds to the regex-interpreted
                      text which will be searched for.
                    * `-i`, `--ignore-case`: if specified, the search will be case *in*sensitive.
                      You can enable this setting permanently with
                      `config.commands.search.ignore_case`.
                    * `-c`, `--context`: you can specify the number of lines of "context" which
                      is the number of lines before and after the actual match to be included in the
                      output. This is similar to `grep`s `-C` option.
                    * in addition to the above, you can add `filters` to narrow the search down to a
                      subset of your database. For more information refer to `cobib.commands.list`.
            out: the output IO stream. This defaults to `None`.

        Returns:
            A tuple containing the number of hits and matching labels.
        """
        LOGGER.debug("Starting Search command.")
        parser = ArgumentParser(prog="search", description="Search subcommand parser.")
        parser.add_argument("query", type=str, help="text to search for")
        parser.add_argument(
            "-i", "--ignore-case", action="store_true", help="ignore case for searching"
        )
        parser.add_argument(
            "-c",
            "--context",
            type=int,
            default=1,
            help="number of context lines to provide for each match",
        )
        parser.add_argument(
            "filter",
            nargs="*",
            help="You can specify filters as used by the `list` command in order to select a "
            "subset of labels to be modified. To ensure this works as expected you should add the "
            "pseudo-argument '--' before the list of filters. See also `list --help` for more "
            "information.",
        )

        if not args:
            parser.print_usage(sys.stderr)
            sys.exit(1)

        try:
            largs = parser.parse_intermixed_args(args)
        except argparse.ArgumentError as exc:
            LOGGER.error(exc.message)
            return None

        Event.PreSearchCommand.fire(largs)

        with open(os.devnull, "w", encoding="utf-8") as devnull:
            labels = ListCommand().execute(largs.filter, out=devnull)
        if labels is None:
            return None  # pragma: no cover
        LOGGER.debug("Available entries to search: %s", labels)

        ignore_case = config.commands.search.ignore_case or largs.ignore_case
        re_flags = re.IGNORECASE if ignore_case else 0
        LOGGER.debug("The search will be performed case %ssensitive", "in" if ignore_case else "")

        bib = Database()

        hits = 0
        output = []
        for label in labels.copy():
            entry = bib[label]
            matches = entry.search(largs.query, largs.context, ignore_case)
            if not matches:
                labels.remove(label)
                continue

            hits += len(matches)
            LOGGER.debug('Entry "%s" includes %d hits.', label, hits)
            title = f"{label} - {len(matches)} match" + ("es" if len(matches) > 1 else "")
            title = title.replace(label, config.get_ansi_color("search_label") + label + "\x1b[0m")
            output.append(title)

            for idx, match in enumerate(matches):
                for line in match:
                    line = re.sub(
                        rf"({largs.query})",
                        config.get_ansi_color("search_query") + r"\1" + "\x1b[0m",
                        line,
                        flags=re_flags,
                    )
                    output.append(f"[{idx+1}]\t".expandtabs(8) + line)

        print("\n".join(output), file=out)

        Event.PostSearchCommand.fire(hits, labels)

        return (hits, labels)

    @staticmethod
    def tui(tui: cobib.tui.TUI) -> None:
        # pdoc will inherit the docstring from the base class
        # noqa: D102
        LOGGER.debug("Search command triggered from TUI.")
        tui.viewport.clear()
        # handle input via prompt
        command, results = tui.execute_command("search", out=tui.viewport.buffer)  # type: ignore
        if tui.viewport.buffer.lines and results is not None:
            hits, labels = results
            tui.STATE.mode = "search"
            cur_y, _ = tui.viewport.pad.getyx()
            tui.STATE.previous_line = cur_y
            tui.viewport.buffer.split()
            LOGGER.debug("Applying selection highlighting in search results.")
            for label in labels:
                if label not in tui.selection:
                    continue
                # we match the label including its 'search_label' highlight to ensure that we really
                # only match this specific occurrence of whatever the label may be
                tui.viewport.buffer.replace(
                    range(tui.viewport.buffer.height),
                    re.escape(config.get_ansi_color("search_label")) + label + re.escape("\x1b[0m"),
                    config.get_ansi_color("search_label")
                    + config.get_ansi_color("selection")
                    + label
                    + "\x1b[0m\x1b[0m",
                )
            LOGGER.debug("Populating viewport with search results.")
            tui.viewport.view(ansi_map=tui.ANSI_MAP)
            # reset current cursor position
            LOGGER.debug("Resetting cursor position to top.")
            tui.STATE.top_line = 0
            tui.STATE.current_line = 0
            # update top statusbar
            tui.STATE.topstatus = f"coBib v{__version__} - {hits} hit{'s' if hits > 1 else ''}"
            tui.statusbar(tui.topbar, tui.STATE.topstatus)
            tui.STATE.inactive_commands = ["Add", "Filter", "Sort"]
        elif command[1:]:
            if sys.version_info[1] >= 8:
                joined_command = shlex.join(command[1:])
            else:
                joined_command = shlex.quote(" ".join(command[1:]))
            msg = f"No search hits for '{joined_command}'!"
            LOGGER.info(msg)
            tui.prompt_print(msg)
            tui.viewport.update_list()



"""coBib's Export command.

You can use this command to export your database.
As of now only two output formats are available:
* BibLaTex files
* Zip archives

The use case of the former is obvious and can be achieved by the following:
```
cobib export --bibtex my_database.bib
```
The latter case will collect all associated files of your database in a single Zip archive:
```
cobib export --zip my_references.zip
```
This is an important feature because coBib (by design) allows you to spread associated files across
your entire file system.
With this command you can gather them in a neat package for sharing or transferring.

You can also limit the export to a subset of your database in one of two ways:
1. through filters:
```
cobib export --bibtex my_private_database.bib -- ++tags private
```
2. through a custom selection (using `--selection`)
```
cobib export --selection --bibtex some_other_database.bib -- Label1 Label2
```
While this latter case is usable via the command-line interface it is more a side-effect of the TUI
integration which provides a visual selection (defaults to the `v` key).
The proper and arguably more useful case is the first case using filters.

Since v3.2.0, coBib supports automatic Journal abbreviations. After configuring them as explained in
`config.utils.journal_abbreviations` you can leverage them during exporting like so:
```
cobib export --abbreviate --bibtex my_database.bib
# or
cobib export --abbreviate --dotless --bibtex my_database.bib
```

You can also trigger this command from the `cobib.tui.tui.TUI`.
By default, it is bound to the `x` key which will drop you into the prompt where you can type out a
normal command-line command:
```
:export <arguments go here>
```
"""

from __future__ import annotations

import argparse
import logging
import os
import sys
from typing import IO, TYPE_CHECKING, Any, List
from zipfile import ZipFile

from cobib.config import Event
from cobib.database import Database
from cobib.parsers.bibtex import BibtexParser
from cobib.utils.journal_abbreviations import JournalAbbreviations
from cobib.utils.rel_path import RelPath

from .base_command import ArgumentParser, Command
from .list import ListCommand

LOGGER = logging.getLogger(__name__)

if TYPE_CHECKING:
    import cobib.tui


class ExportCommand(Command):
    """The Export Command."""

    name = "export"

    def execute(self, args: List[str], out: IO[Any] = sys.stdout) -> None:
        """Exports the database.

        This command exports the database (or a selected subset of entries).
        You can choose the exported formats from the following list:
        * BibLaTex (via the `--bibtex` argument)
        * Zip archive (via the `--zip` argument)

        Args:
            args: a sequence of additional arguments used for the execution. The following values
                are allowed for this command:
                    * `-b`, `--bibtex`: specifies a BibLaTex filename into which to export.
                    * `-z`, `--zip`: specifies a Zip-filename into which to export associated files.
                    * `-a`, `--abbreviate`: abbreviate the Journal names before exporting. See also
                      `config.utils.journal_abbreviations`.
                    * `--dotless`: remove punctuation from the Journal abbreviations.
                    * `-s`, `--selection`: when specified, the positional arguments will *not* be
                      interpreted as filters but rather as a direct list of entry labels. This can
                      be used on the command-line but is mainly meant for the TUIs visual selection
                      interface (hence the name).
                    * in addition to the above, you can add `filters` to specify a subset of your
                      database for exporting. For more information refer to `cobib.commands.list`.
            out: the output IO stream. This defaults to `sys.stdout`.
        """
        LOGGER.debug("Starting Export command.")
        parser = ArgumentParser(prog="export", description="Export subcommand parser.")
        parser.add_argument(
            "-b", "--bibtex", type=argparse.FileType("a"), help="BibLaTeX output file"
        )
        parser.add_argument("-z", "--zip", type=argparse.FileType("a"), help="zip output file")
        parser.add_argument(
            "-s",
            "--selection",
            action="store_true",
            help="When specified, the `filter` argument will be interpreted as a list of entry "
            "labels rather than arguments for the `list` command.",
        )
        parser.add_argument(
            "filter",
            nargs="*",
            help="You can specify filters as used by the `list` command in order to select a "
            "subset of labels to be modified. To ensure this works as expected you should add the "
            "pseudo-argument '--' before the list of filters. See also `list --help` for more "
            "information.",
        )
        parser.add_argument(
            "-a", "--abbreviate", action="store_true", help="Abbreviate journal names"
        )
        parser.add_argument(
            "--dotless", action="store_true", help="Remove punctuation from journal abbreviations"
        )

        if not args:
            parser.print_usage(sys.stderr)
            sys.exit(1)

        try:
            largs = parser.parse_intermixed_args(args)
        except argparse.ArgumentError as exc:
            LOGGER.error(exc.message)
            return

        Event.PreExportCommand.fire(largs)

        if largs.bibtex is None and largs.zip is None:
            msg = "No output file specified!"
            LOGGER.error(msg)
            return
        if largs.zip is not None:
            largs.zip = ZipFile(largs.zip.name, "w")  # pylint: disable=consider-using-with

        if largs.selection:
            LOGGER.info("Selection given. Interpreting `filter` as a list of labels")
            labels = largs.filter
        else:
            LOGGER.debug("Gathering filtered list of entries to be exported.")
            with open(os.devnull, "w", encoding="utf-8") as devnull:
                labels = ListCommand().execute(largs.filter, out=devnull)

        bibtex_parser = BibtexParser()

        bib = Database()

        for label in labels:
            try:
                LOGGER.info('Exporting entry "%s".', label)
                entry = bib[label]
                if largs.bibtex is not None:
                    if largs.abbreviate and "journal" in entry.data.keys():
                        entry.data["journal"] = JournalAbbreviations.abbreviate(
                            entry.data["journal"], dotless=largs.dotless
                        )
                    entry_str = bibtex_parser.dump(entry)
                    largs.bibtex.write(entry_str)
                if largs.zip is not None:
                    if "file" in entry.data.keys() and entry.file is not None:
                        files = entry.file
                        if not isinstance(files, list):
                            files = [files]
                        for file in files:
                            path = RelPath(file).path
                            LOGGER.debug(
                                'Adding "%s" associated with "%s" to the zip file.', path, label
                            )
                            largs.zip.write(path, path.name)
            except KeyError:
                msg = f"No entry with the label '{label}' could be found."
                LOGGER.warning(msg)

        Event.PostExportCommand.fire(labels, largs)

        if largs.zip is not None:
            largs.zip.close()

    @staticmethod
    def tui(tui: cobib.tui.TUI) -> None:
        # pdoc will inherit the docstring from the base class
        # noqa: D102
        LOGGER.debug("Export command triggered from TUI.")
        # handle input via prompt
        if tui.selection:
            tui.execute_command("export -s", pass_selection=True)
        else:
            tui.execute_command("export")



"""coBib's Delete command.

This command can be used to deleted entries from the database.
```
cobib delete <label 1> [<label 2> ...]
```

You can also trigger this command from the `cobib.tui.tui.TUI`.
By default, it is bound to the `d` key.
"""

from __future__ import annotations

import argparse
import logging
import os
import sys
from typing import IO, TYPE_CHECKING, Any, List

from cobib.config import Event
from cobib.database import Database
from cobib.utils.rel_path import RelPath

from .base_command import ArgumentParser, Command

LOGGER = logging.getLogger(__name__)

if TYPE_CHECKING:
    import cobib.tui


class DeleteCommand(Command):
    """The Delete Command."""

    name = "delete"

    def execute(self, args: List[str], out: IO[Any] = sys.stdout) -> None:
        """Deletes an entry.

        This command deletes one (or multiple) entries from the database.

        Args:
            args: a sequence of additional arguments used for the execution. The following values
                are allowed for this command:
                    * `labels`: one (or multiple) labels of the entries to be deleted.
            out: the output IO stream. This defaults to `sys.stdout`.
        """
        LOGGER.debug("Starting Delete command.")
        parser = ArgumentParser(prog="delete", description="Delete subcommand parser.")
        parser.add_argument("labels", type=str, nargs="+", help="labels of the entries")
        parser.add_argument(
            "--preserve-files", action="store_true", help="do not delete associated files"
        )

        if not args:
            parser.print_usage(sys.stderr)
            sys.exit(1)

        try:
            largs = parser.parse_args(args)
        except argparse.ArgumentError as exc:
            LOGGER.error(exc.message)
            return

        Event.PreDeleteCommand.fire(largs)

        deleted_entries = set()

        bib = Database()
        for label in largs.labels:
            try:
                LOGGER.debug("Attempting to delete entry '%s'.", label)
                entry = bib.pop(label)
                if not largs.preserve_files:
                    for file in entry.file:
                        path = RelPath(file)
                        try:
                            LOGGER.debug("Attempting to remove associated file '%s'.", str(path))
                            os.remove(path.path)
                        except FileNotFoundError:
                            pass

                deleted_entries.add(label)
            except KeyError:
                pass

        Event.PostDeleteCommand.fire(deleted_entries)
        bib.save()

        self.git(args=vars(largs))

        for label in deleted_entries:
            msg = f"'{label}' was removed from the database."
            LOGGER.info(msg)

    @staticmethod
    def tui(tui: cobib.tui.TUI) -> None:
        # pdoc will inherit the docstring from the base class
        # noqa: D102
        LOGGER.debug("Delete command triggered from TUI.")
        if tui.selection:
            # use selection for command
            labels = list(tui.selection)
            tui.selection.clear()
        else:
            # get current label
            label, _ = tui.viewport.get_current_label()
            labels = [label]
        # delete selected entry
        tui.execute_command(["delete"] + labels, skip_prompt=True)
        # update database list
        LOGGER.debug("Updating list after Delete command.")
        tui.viewport.update_list()



"""coBib's Edit command.

This command can be used to manually edit database entries in their easily-readable YAML format.
To get started, simply type:
```
cobib edit <label>
```
which will open the YAML-formatted version of the specified Entry for editing.

You can configure which editor will be used via the `config.commands.edit.editor` setting which will
default to using your `$EDITOR` environment setting (and fall back to `vim` if that is not set).

You can even add entirely new entries to the database by specifying an unused entry label *and*
adding the `--add` command-line argument:
```
cobib edit --add <new label>
```
This entry will be entirely empty except for the one field which is always present:
* `ENTRYTYPE`: set to the default value configured via `config.commands.edit.default_entry_type`.

You can also trigger this command from the `cobib.tui.tui.TUI`.
By default, it is bound to the `e` key.
If you want to add a new entry manually, you will have to enter the prompt (defaults to `:`) and
then type out the command mentioned above:
```
:edit --add <new label>
```
"""

from __future__ import annotations

import argparse
import logging
import os
import sys
import tempfile
from pathlib import Path
from typing import IO, TYPE_CHECKING, Any, List

from cobib.config import Event, config
from cobib.database import Database, Entry
from cobib.parsers.yaml import YAMLParser
from cobib.utils.rel_path import RelPath

from .base_command import ArgumentParser, Command

LOGGER = logging.getLogger(__name__)

if TYPE_CHECKING:
    import cobib.tui


class EditCommand(Command):
    """The Edit Command."""

    name = "edit"

    def execute(self, args: List[str], out: IO[Any] = sys.stdout) -> None:
        """Opens an entry for manual editing.

        This command opens an `cobib.database.Entry` in YAML format for manual editing.
        The editor program can be configured via `config.commands.edit.editor`.
        By default, this setting will respect your `$EDITOR` environment variable, but fall back to
        using `vim` if that variable is not set.

        Args:
            args: a sequence of additional arguments used for the execution. The following values
                are allowed for this command:
                    * `label`: the label of the entry to edit.
                    * `-a`, `--add`: if specified, allows adding new entries for non-existent
                      labels. The default entry type of this new entry can be configured via
                      `config.commands.edit.default_entry_type`.
            out: the output IO stream. This defaults to `sys.stdout`.
        """
        LOGGER.debug("Starting Edit command.")
        parser = ArgumentParser(prog="edit", description="Edit subcommand parser.")
        parser.add_argument("label", type=str, help="label of the entry")
        parser.add_argument(
            "-a",
            "--add",
            action="store_true",
            help="if specified, will add a new entry for unknown labels",
        )
        parser.add_argument(
            "--preserve-files", action="store_true", help="do not rename associated files"
        )

        if not args:
            parser.print_usage(sys.stderr)
            sys.exit(1)

        try:
            largs = parser.parse_args(args)
        except argparse.ArgumentError as exc:
            LOGGER.error(exc.message)
            return

        Event.PreEditCommand.fire(largs)

        yml = YAMLParser()

        bib = Database()

        try:
            entry = bib[largs.label]
            prv = yml.dump(entry)
            if largs.add:
                LOGGER.warning(
                    "Entry '%s' already exists! Ignoring the `--add` argument.", largs.label
                )
                largs.add = False
        except KeyError:
            # No entry for given label found
            if largs.add:
                # add a new entry for the unknown label
                entry = Entry(
                    largs.label,
                    {"ENTRYTYPE": config.commands.edit.default_entry_type},
                )
                prv = yml.dump(entry)
            else:
                msg = (
                    f"No entry with the label '{largs.label}' could be found."
                    "\nUse `--add` to add a new entry with this label."
                )
                LOGGER.error(msg)
                return
        if prv is None:
            # No entry found to be edited. This should never occur unless the YAMLParser experiences
            # an unexpected error.
            return

        LOGGER.debug("Creating temporary file.")
        with tempfile.NamedTemporaryFile(mode="w+", prefix="cobib-", suffix=".yaml") as tmp_file:
            tmp_file_name = tmp_file.name
            tmp_file.write(prv)
            tmp_file.flush()
            LOGGER.debug('Starting editor "%s".', config.commands.edit.editor)
            status = os.system(config.commands.edit.editor + " " + tmp_file.name)
            assert status == 0
            LOGGER.debug("Editor finished successfully.")
            new_entries = YAMLParser().parse(tmp_file.name)
            new_entry = list(new_entries.values())[0]
        assert not Path(tmp_file_name).exists()
        if entry == new_entry and not largs.add:
            LOGGER.info("No changes detected.")
            return

        bib.update({new_entry.label: new_entry})
        if new_entry.label != largs.label:
            bib.rename(largs.label, new_entry.label)
            if not largs.preserve_files:
                new_files = []
                for file in new_entry.file:
                    path = RelPath(file)
                    if path.path.stem == largs.label:
                        LOGGER.info("Also renaming associated file '%s'.", str(path))
                        target = RelPath(path.path.parent / f"{new_entry.label}.pdf")
                        if target.path.exists():
                            LOGGER.warning("Found conflicting file, not renaming '%s'.", str(path))
                        else:
                            path.path.rename(target.path)
                            new_files.append(str(target))
                            continue
                    new_files.append(file)
                new_entry.file = new_files

        Event.PostEditCommand.fire(new_entry)
        bib.save()

        self.git(args=vars(largs))

        msg = f"'{largs.label}' was successfully edited."
        LOGGER.info(msg)

    @staticmethod
    def tui(tui: cobib.tui.TUI) -> None:
        # pdoc will inherit the docstring from the base class
        # noqa: D102
        LOGGER.debug("Edit command triggered from TUI.")
        # get current label
        label, _ = tui.viewport.get_current_label()
        # populate buffer with entry data
        EditCommand().execute([label])
        # redraw total screen after closing external editor
        LOGGER.debug("Manually redrawing TUI to clear out any editor artifacts.")
        tui.resize_handler(None, None)
        # update database list
        tui.viewport.update_list()



"""coBib's Add command.

This command allows you to add new entries to your database.
You have two options on how to use this command.

### 1. Parser-based addition

All parsers available in `cobib.parsers` will be registered (at runtime) in a *mutually exclusive*
group of keyword arguments.
This means, that you can add a new entry based on any of them by using the keyword argument
according to the parsers name.
For example:
```
cobib add --bibtex some_biblatex_file.bib
cobib add --arxiv <some arXiv ID>
cobib add --doi <some DOI>
cobib add --isbn <some ISBN>
cobib add --url <some URL>
cobib add --yaml some_cobib_style_yaml_file.yaml
```

Note, that you cannot combine multiple parsers within a single command execution (for obvious
reasons).

Furthermore, most of these parsers will set the entry label automatically.
If you would like to prevent that and specify a custom label directly, you can add the `--label`
keyword argument to your command like so:
```
cobib add --doi <some DOI> --label "MyLabel"
```

### 2. Manual entry addition

If you want to add a new entry manually, you *must* omit any of the parser keyword arguments and
instead specify a new label like so:
```
cobib add --label <some new label>
```
This will trigger the `cobib.commands.edit.EditCommand` for a manual addition.
However, the benefit of using this through the `AddCommand` is the availability of the following
additional options which are always available (i.e. also in combination with the parser keyword
arguments, above).

### Additional Options

You can directly associate your new entry with one or multiple files by doing the following:
```
cobib add --doi <some DOI> --file /path/to/a/file.pdf [/path/to/another/file.dat ...]
```

You can also specify `cobib.database.Entry.tags` using *positional* arguments like so:
```
cobib add --doi <some DOI> -- tag1 "multi-word tag2" ...
```

Since v3.2.0 coBib attempts to download PDF files for newly added entries (if the corresponding
parser supports this feature). The default location where this file will be stored can be configured
via `config.utils.file_downloader.default_location`, but it can be changed at runtime using the
`--path` argument like so:
```
cobib add --path <some custom path> --arxiv <some arXiv ID>
```
If you want to manually suppress the automatic download, specify the `--skip-download` argument:
```
cobib add --skip-download --arxiv <some arXiv ID>
```

Since v3.3.0 you can also update existing entries by using the `--update` argument:
```
cobib add --doi <some DOI> --update --label <some existing label>
```
This will take the existing entry and combine it with all new information found in the freshly added
entry. Existing fields will be overwritten. If you have an automatically downloaded file associated
with this entry, that will also be overwritten.
This feature is especially useful if you want to update an entry which you previously added from the
arXiv with its newly published version.

If you don't specify `--update` and run into a situation where the label which you are trying to add
already exists, coBib will disambiguate it based on the `config.database.format.label_suffix`
setting. It defaults to appending `_a`, `_b`, etc.
You can disable this disambiguation by passing `--skip-existing` to the add command.

### TUI

You can also trigger this command from the `cobib.tui.tui.TUI`.
By default, it is bound to the `a` key which will drop you into the prompt where you can type out a
normal command-line command:
```
:add <arguments go here>
```
"""

from __future__ import annotations

import argparse
import inspect
import logging
import sys
from collections import OrderedDict
from typing import IO, TYPE_CHECKING, Any, Dict, List

from cobib import parsers
from cobib.config import Event, config
from cobib.database import Database, Entry
from cobib.utils.file_downloader import FileDownloader
from cobib.utils.journal_abbreviations import JournalAbbreviations

from .base_command import ArgumentParser, Command
from .edit import EditCommand
from .modify import evaluate_as_f_string

LOGGER = logging.getLogger(__name__)

if TYPE_CHECKING:
    import cobib.tui


class AddCommand(Command):
    """The Add Command."""

    name = "add"

    # pylint: disable=too-many-branches,too-many-statements
    def execute(self, args: List[str], out: IO[Any] = sys.stdout) -> None:
        """Adds a new entry.

        Depending on the `args`, if a keyword for one of the available `cobib.parsers` was used
        together with a matching input, that parser will be used to create the new entry.
        Otherwise, the command is only valid if the `--label` option was used to specify a new entry
        label, in which case this command will trigger the `cobib.commands.edit.EditCommand` for a
        manual entry addition.

        Args:
            args: a sequence of additional arguments used for the execution. The following values
                are allowed for this command:
                    * `-l`, `--label`: the label to give to the new entry.
                    * `-u`, `--update`: updates an existing database entry if it already exists.
                    * `-f`, `--file`: one or multiple files to associate with this entry. This data
                      will be stored in the `cobib.database.Entry.file` property.
                    * `-p`, `--path`: the path to store the downloaded associated file in. This can
                      be used to overwrite the `config.utils.file_downloader.default_location`.
                    * `--skip-download`: skips the automatic download of an associated file.
                    * `--skip-existing`: skips entry if label exists instead of running label
                      disambiguation.
                    * in addition to the options above, a *mutually exclusive group* of keyword
                      arguments for all available `cobib.parsers` are registered at runtime. Please
                      check the output of `cobib add --help` for the exact list.
                    * any *positional* arguments (i.e. those, not preceded by a keyword) are
                      interpreted as tags and will be stored in the `cobib.database.Entry.tags`
                      property.
            out: the output IO stream. This defaults to `sys.stdout`.
        """
        LOGGER.debug("Starting Add command.")
        parser = ArgumentParser(prog="add", description="Add subcommand parser.")
        parser.add_argument("-l", "--label", type=str, help="the label for the new database entry")
        parser.add_argument(
            "-u",
            "--update",
            action="store_true",
            help="update an entry if the label exists already",
        )
        file_action = "extend" if sys.version_info[1] >= 8 else "append"
        parser.add_argument(
            "-f",
            "--file",
            type=str,
            nargs="+",
            action=file_action,
            help="files associated with this entry",
        )
        parser.add_argument("-p", "--path", type=str, help="the path for the associated file")
        parser.add_argument(
            "--skip-download",
            action="store_true",
            help="skip the automatic download of an associated file",
        )
        parser.add_argument(
            "--skip-existing",
            action="store_true",
            help="skips entry addition if existent instead of using label disambiguation",
        )
        group_add = parser.add_mutually_exclusive_group()
        avail_parsers = {
            cls.name: cls for _, cls in inspect.getmembers(parsers) if inspect.isclass(cls)
        }
        for name in avail_parsers.keys():
            try:
                group_add.add_argument(
                    f"-{name[0]}", f"--{name}", type=str, help=f"{name} object identfier"
                )
            except argparse.ArgumentError:
                try:
                    group_add.add_argument(f"--{name}", type=str, help=f"{name} object identfier")
                except argparse.ArgumentError:
                    continue
        parser.add_argument(
            "tags",
            nargs=argparse.REMAINDER,
            help="A list of space-separated tags to associate with this entry."
            "\nYou can use quotes to specify tags with spaces in them.",
        )
        if not args:
            parser.print_usage(sys.stderr)
            sys.exit(1)

        try:
            largs = parser.parse_args(args)
        except argparse.ArgumentError as exc:
            LOGGER.error(exc.message)
            return

        Event.PreAddCommand.fire(largs)

        new_entries: Dict[str, Entry] = OrderedDict()

        edit_entries = False
        for name, cls in avail_parsers.items():
            string = getattr(largs, name, None)
            if string is None:
                continue
            LOGGER.debug("Adding entries from %s: '%s'.", name, string)
            new_entries = cls().parse(string)
            break
        else:
            if largs.label is not None:
                LOGGER.warning("No input to parse. Creating new entry '%s' manually.", largs.label)
                new_entries = {
                    largs.label: Entry(
                        largs.label,
                        {"ENTRYTYPE": config.commands.edit.default_entry_type},
                    )
                }
                edit_entries = True
            else:
                msg = "Neither an input to parse nor a label for manual creation specified!"
                LOGGER.error(msg)
                return

        if largs.label is not None:
            assert len(new_entries.values()) == 1
            for value in new_entries.values():
                # logging done by cobib/database/entry.py
                value.label = largs.label
            new_entries = OrderedDict((largs.label, value) for value in new_entries.values())
        else:
            formatted_entries = OrderedDict()
            for label, value in new_entries.items():
                formatted_label = evaluate_as_f_string(
                    config.database.format.label_default, {"label": label, **value.data.copy()}
                )
                value.label = formatted_label
                formatted_entries[formatted_label] = value
            new_entries = formatted_entries

        if largs.file is not None:
            if file_action == "append":
                # We need to flatten the potentially nested list.
                # pylint: disable=import-outside-toplevel
                from itertools import chain

                largs.file = list(chain.from_iterable(largs.file))
            assert len(new_entries.values()) == 1
            for value in new_entries.values():
                # logging done by cobib/database/entry.py
                value.file = largs.file

        if largs.tags != []:
            assert len(new_entries.values()) == 1
            for value in new_entries.values():
                # logging done by cobib/database/entry.py
                value.tags = largs.tags

        bib = Database()
        existing_labels = set(bib.keys())

        for lbl, entry in new_entries.copy().items():
            # check if label already exists
            if lbl in existing_labels:
                if not largs.update:
                    msg = f"You tried to add a new entry '{lbl}' which already exists!"
                    LOGGER.warning(msg)
                    if edit_entries or largs.skip_existing:
                        msg = f"Please use `cobib edit {lbl}` instead!"
                        LOGGER.warning(msg)
                        continue
                    msg = (
                        "The label will be disambiguated based on the configuration option: "
                        "config.database.format.label_suffix"
                    )
                    LOGGER.warning(msg)
                    new_label = bib.disambiguate_label(lbl)
                    entry.label = new_label
                    new_entries[new_label] = entry
                    new_entries.pop(lbl)
                else:
                    # label exists but the user asked to update an existing entry
                    existing_data = bib[lbl].data.copy()
                    existing_data.update(entry.data)
                    entry.data = existing_data.copy()
            # download associated file (if requested)
            if "_download" in entry.data.keys():
                if largs.skip_download:
                    entry.data.pop("_download")
                else:
                    path = FileDownloader().download(
                        entry.data.pop("_download"), lbl, folder=largs.path, overwrite=largs.update
                    )
                    if path is not None:
                        entry.data["file"] = str(path)
            # check journal abbreviation
            if "journal" in entry.data.keys():
                entry.data["journal"] = JournalAbbreviations.elongate(entry.data["journal"])

        Event.PostAddCommand.fire(new_entries)

        bib.update(new_entries)
        if edit_entries:
            EditCommand().execute([largs.label])

        bib.save()

        self.git(args=vars(largs))

        for label in new_entries:
            msg = f"'{label}' was added to the database."
            LOGGER.info(msg)

    @staticmethod
    def tui(tui: cobib.tui.TUI) -> None:
        # pdoc will inherit the docstring from the base class
        # noqa: D102
        LOGGER.debug("Add command triggered from TUI.")
        # handle input via prompt
        tui.execute_command("add")
        # update database list
        LOGGER.debug("Updating list after Add command.")
        tui.viewport.update_list()
